# Copyright Contributors to the osc project.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.


"""
This module handles configuration of osc.


Configuring osc from oscrc
--------------------------

To configure osc from oscrc, do following::

    import osc.conf

    # see ``get_config()`` documentation for available function arguments
    # see ``oscrc(5)`` man page for oscrc configuration options
    osc.conf.get_config()


Configuring osc from API
------------------------

To configure osc purely from the API (without reading oscrc), do following::

    import osc.conf

    # initialize the main config object
    config = osc.conf.Options()

    # configure host options for an apiurl
    apiurl = osc.conf.sanitize_apiurl(apiurl)
    host_options = HostOptions(apiurl=apiurl, username=..., _parent=config)
    config.api_host_options[apiurl] = host_options

    # set the default ``apiurl``
    config.apiurl = ...

    # place the config object in `osc.conf`
    osc.conf.config = config

    # optional: enable http debugging according to the ``http_debug`` and ``http_full_debug`` options
    from osc.connection import enable_http_debug
    enable_http_debug(osc.conf.config)
"""


import collections
import errno
import getpass
import http.client
import os
import re
import shutil
import sys
import textwrap
from io import BytesIO
from io import StringIO
from urllib.parse import urlsplit

from . import credentials
from . import OscConfigParser
from . import oscerr
from .output import tty
from .util import xdg
from .util.helper import raw_input
from .util.models import *


GENERIC_KEYRING = False

try:
    import keyring
    GENERIC_KEYRING = True
except:
    pass


__all__ = [
    "get_config",
    "Options",
    "HostOptions",
    "Password",
    "config",
]


class Password(collections.UserString):
    """
    Lazy password that wraps either a string or a function.
    The result of the function gets returned any time the object is used as a string.
    """

    def __init__(self, data):
        self._data = data

    @property
    def data(self):
        if callable(self._data):
            # if ``data`` is a function, call it every time the string gets evaluated
            # we use the password only from time to time to make a session cookie
            # and there's no need to keep the password in program memory longer than necessary
            result = self._data()

            # the function can also return a function, let's evaluate them recursively
            while callable(result):
                result = result()

            if result is None:
                raise oscerr.OscIOError(None, "Unable to retrieve password")
            return result
        return self._data

    def __format__(self, format_spec):
        if format_spec.endswith("s"):
            return f"{self.__str__():{format_spec}}"
        return super().__format__(format_spec)

    def encode(self, *args, **kwargs):
        if sys.version_info < (3, 8):
            # avoid returning the Password object on python < 3.8
            return str(self).encode(*args, **kwargs)
        return super().encode(*args, **kwargs)


HttpHeader = NewType("HttpHeader", Tuple[str, str])


class OscOptions(BaseModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_new_attributes = True
        self._extra_fields = {}
        self._allow_new_attributes = False

    # compat function with the config dict
    def _get_field_name(self, name):
        if name in self.__fields__:
            return name

        for field_name, field in self.__fields__.items():
            ini_key = field.extra.get("ini_key", None)
            if ini_key == name:
                return field_name

        return None

    # compat function with the config dict
    def __getitem__(self, name):
        field_name = self._get_field_name(name)

        if field_name is None and not hasattr(self, name):
            return self._extra_fields[name]

        field_name = field_name or name
        try:
            return getattr(self, field_name)
        except AttributeError:
            raise KeyError(name)

    # compat function with the config dict
    def __setitem__(self, name, value):
        field_name = self._get_field_name(name)

        if field_name is None and not hasattr(self, name):
            self._extra_fields[name] = value
            return

        field_name = field_name or name
        setattr(self, field_name, value)

    # compat function with the config dict
    def __contains__(self, name):
        try:
            self[name]
        except KeyError:
            return False
        return True

    # compat function with the config dict
    def setdefault(self, name, default=None):
        field_name = self._get_field_name(name)
        # we're ignoring ``default`` because the field always exists
        return getattr(self, field_name, None)

    # compat function with the config dict
    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def set_value_from_string(self, name, value):
        field_name = self._get_field_name(name)
        field = self.__fields__[field_name]

        if not isinstance(value, str):
            setattr(self, field_name, value)
            return

        if not value.strip():
            if field.is_optional:
                setattr(self, field_name, None)
                return

        if field.origin_type is Password:
            value = Password(value)
            setattr(self, field_name, value)
            return

        if field.type is List[HttpHeader]:
            value = http.client.parse_headers(BytesIO(value.strip().encode("utf-8"))).items()
            setattr(self, field_name, value)
            return

        if field.origin_type is list:
            # split list options into actual lists
            value = re.split(r"[, ]+", value)
            setattr(self, field_name, value)
            return

        if field.origin_type is bool:
            if value.lower() in ["1", "yes", "true", "on"]:
                value = True
                setattr(self, field_name, value)
                return
            if value.lower() in ["0", "no", "false", "off"]:
                value = False
                setattr(self, field_name, value)
                return

        if field.origin_type is int:
            value = int(value)
            setattr(self, field_name, value)
            return

        setattr(self, field_name, value)


class HostOptions(OscOptions):
    """
    Configuration options for individual apiurls.
    """

    def __init__(self, _parent, **kwargs):
        super().__init__(_parent=_parent, **kwargs)

    apiurl: str = Field(
        default=None,
        description=textwrap.dedent(
            """
            URL to the API server.
            """
        ),
    )  # type: ignore[assignment]

    aliases: List[str] = Field(
        default=[],
        description=textwrap.dedent(
            """
            Aliases of the apiurl.
            """
        ),
    )  # type: ignore[assignment]

    username: str = Field(
        default=None,
        description=textwrap.dedent(
            """
            Username for the apiurl.
            """
        ),
        ini_key="user",
    )  # type: ignore[assignment]

    credentials_mgr_class: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Fully qualified name of a class used to fetch a password.
            """
        ),
    )  # type: ignore[assignment]

    password: Optional[Password] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Password for the apiurl.
            May be empty if the credentials manager fetches the password from a keyring or ``sshkey`` is used.
            """
        ),
        ini_key="pass",
    )  # type: ignore[assignment]

    sshkey: Optional[str] = Field(
        default=FromParent("sshkey"),
        description=textwrap.dedent(
            """
            A pointer to public SSH key that corresponds with a private SSH used for authentication:

             - keep empty for auto detection
             - path to the public SSH key
             - public SSH key filename (must be placed in ~/.ssh)
             - fingerprint of a SSH key (2nd column of ``ssh-add -l``)

            NOTE: The private key may not be available on disk because it could be in a GPG keyring, on YubiKey or forwarded through SSH agent.

            TIP: To give osc a hint which ssh key from the agent to use during auto detection,
            append ``obs=<apiurl-hostname>`` to the **private** key's comment.
            This will also work nicely during SSH agent forwarding, because the comments get forwarded too.

             - To edit the key, run: ``ssh-keygen -c -f ~/.ssh/<private-key>``
             - To query the key, run: ``ssh-keygen -y -f ~/.ssh/<private-key>``
             - Example comment: ``<username@host> obs=api.example.com obs=api-test.example.com``
            """
        ),
    )  # type: ignore[assignment]

    downloadurl: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Redirect downloads of packages used during build to an alternative location.
            This allows specifying a local mirror or a proxy, which can greatly improve download performance, latency and more.
            """
        ),
    )  # type: ignore[assignment]

    cafile: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The path to a file of concatenated CA certificates in PEM format.
            If specified, the CA certificates from the path will be used to validate other peers' certificates instead of the system-wide certificates.
            """
        ),
    )  # type: ignore[assignment]

    capath: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The path to a directory containing several CA certificates in PEM format.
            If specified, the CA certificates from the path will be used to validate other peers' certificates instead of the system-wide certificates.
            """
        ),
    )  # type: ignore[assignment]

    sslcertck: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Whether to validate SSL certificate of the server.
            It is highly recommended to keep this option enabled.
            """
        ),
    )  # type: ignore[assignment]

    allow_http: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Whether to allow plain HTTP connections.
            Using HTTP is insecure because it sends passwords and session cookies in plain text.
            It is highly recommended to keep this option disabled.
            """
        ),
    )  # type: ignore[assignment]

    http_headers: List[HttpHeader] = Field(
        default=[],
        description=textwrap.dedent(
            """
            Additional HTTP headers attached to each HTTP or HTTPS request.
            The format is [(header-name, header-value)].
            """
        ),
        ini_description=textwrap.dedent(
            """
            Additional HTTP headers attached to each request.
            The format is HTTP headers separated with newlines.

            Example::

                http_headers =
                    X-Header1: Value1
                    X-Header2: Value2
            """
        ),
        ini_type="newline-separated-list",
    )  # type: ignore[assignment]

    trusted_prj: List[str] = Field(
        default=[],
        description=textwrap.dedent(
            """
            List of names of the trusted projects.
            The names can contain globs.
            Please note that some repos may contain malicious packages that can compromise the build result or even your system!
            """
        ),
    )  # type: ignore[assignment]

    disable_hdrmd5_check: bool = Field(
        default=FromParent("disable_hdrmd5_check"),
        description=textwrap.dedent(
            """
            Disable hdrmd5 checks of downloaded and cached packages in ``osc build``.
            It is recommended to keep the check enabled.

            OBS builds the noarch packages once per binary arch.
            Such noarch packages are supposed to be nearly identical across all build arches,
            any discrepancy in the payload and dependencies is considered a packaging bug.
            But to guarantee that the local builds work identically to builds in OBS,
            using the arch-specific copy of the noarch package is required.
            Unfortunatelly only one of the noarch packages gets distributed
            and can be downloaded from a local mirror.
            All other noarch packages are available through the OBS API only.
            Since there is currently no information about hdrmd5 checksums of published noarch packages,
            we download them, verify hdrmd5 and re-download the package from OBS API on mismatch.

            The same can also happen for architecture depend packages when someone is messing around
            with the source history or the release number handling in a way that it is not increasing.

            If you want to save some bandwidth and don't care about the exact rebuilds
            you can turn this option on to disable hdrmd5 checks completely.
            """
        ),
    )  # type: ignore[assignment]

    passx: Optional[str] = Field(
        default=None,
        deprecated_text=textwrap.dedent(
            """
            Option 'passx' (oscrc option [$apiurl]/passx) is deprecated.
            You should be using the 'password' option with 'credentials_mgr_class' set to 'osc.credentials.ObfuscatedConfigFileCredentialsManager' instead.
            """
        ),
    )  # type: ignore[assignment]

    realname: Optional[str] = Field(
        default=FromParent("realname"),
        description=textwrap.dedent(
            """
            Name of the user passed to the ``vc`` tool via ``VC_REALNAME`` env variable.
            """
        ),
    )  # type: ignore[assignment]

    email: Optional[str] = Field(
        default=FromParent("email"),
        description=textwrap.dedent(
            """
            Email of the user passed to the ``vc`` tool via ``VC_MAILADDR`` env variable.
            """
        ),
    )  # type: ignore[assignment]


class Options(OscOptions):
    """
    Main configuration options.
    """

    # for internal use
    conffile: Optional[str] = Field(
        default=None,
        exclude=True,
    )  # type: ignore[assignment]

    api_host_options: Dict[str, HostOptions] = Field(
        default={},
        description=textwrap.dedent(
            """
            A dictionary that maps ``apiurl`` to ``HostOptions``.
            """
        ),
        ini_exclude=True,
    )  # type: ignore[assignment]

    @property
    def apiurl_aliases(self):
        """
        Compute and return a dictionary that maps ``alias`` to ``apiurl``.
        """
        result = {}
        for apiurl, opts in self.api_host_options.items():
            result[apiurl] = apiurl
            for alias in opts.aliases:
                result[alias] = apiurl
        return result

    section_generic: str = Field(
        default="Generic options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    apiurl: str = Field(
        default="https://api.opensuse.org",
        description=textwrap.dedent(
            """
            Default URL to the API server.
            Credentials and other ``apiurl`` specific settings must be configured
            in a ``[$apiurl]`` config section or via API in an ``api_host_options`` entry.
            """
        ),
    )  # type: ignore[assignment]

    section_auth: str = Field(
        default="Authentication options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    username: Optional[str] = Field(
        default=None,
        ini_key="user",
        deprecated_text=textwrap.dedent(
            """
            Option 'username' (oscrc option [global]/user) is deprecated.
            You should be using username for each apiurl instead.
            """
        ),
    )  # type: ignore[assignment]

    password: Optional[Password] = Field(
        default=None,
        ini_key="pass",
        deprecated_text=textwrap.dedent(
            """
            Option 'password' (oscrc option [global]/pass) is deprecated.
            You should be using password for each apiurl instead.
            """
        ),
    )  # type: ignore[assignment]

    passx: Optional[str] = Field(
        default=None,
        deprecated_text=textwrap.dedent(
            """
            Option 'passx' (oscrc option [global]/passx) is deprecated.
            You should be using password for each apiurl instead.
            """
        ),
    )  # type: ignore[assignment]

    sshkey: Optional[str] = Field(
        default=None,
        description=HostOptions.__fields__["sshkey"].description,
    )  # type: ignore[assignment]

    use_keyring: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Enable keyring as an option for storing passwords.
            """
        ),
    )  # type: ignore[assignment]

    section_verbosity: str = Field(
        default="Verbosity options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    quiet: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Reduce amount of printed information to bare minimum.
            If enabled, automatically sets ``verbose`` to ``False``.
            """
        ),
    )  # type: ignore[assignment]

    verbose: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Increase amount of printed information to stdout.
            Automatically set to ``False`` when ``quiet`` is enabled.
            """
        ),
        get_callback=lambda conf, value: False if conf.quiet else value,
    )  # type: ignore[assignment]

    debug: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Print debug information to stderr.
            """
        ),
    )  # type: ignore[assignment]

    http_debug: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Print HTTP traffic to stderr.
            Automatically set to ``True`` when``http_full_debug`` is enabled.
            """
        ),
        get_callback=lambda conf, value: True if conf.http_full_debug else value,
    )  # type: ignore[assignment]

    http_full_debug: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            [CAUTION!] Print HTTP traffic incl. authentication data to stderr.
            If enabled, automatically sets ``http_debug`` to ``True``.
            """
        ),
    )  # type: ignore[assignment]

    post_mortem: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Jump into a debugger when an unandled exception occurs.
            """
        ),
    )  # type: ignore[assignment]

    traceback: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Print full traceback to stderr when an unandled exception occurs.
            """
        ),
    )  # type: ignore[assignment]

    show_download_progress: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Show download progressbar.
            """
        ),
    )  # type: ignore[assignment]

    section_connection: str = Field(
        default="Connection options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    http_retries: int = Field(
        default=3,
        description=textwrap.dedent(
            """
            Number of retries on HTTP error.
            """
        ),
    )  # type: ignore[assignment]

    cookiejar: str = Field(
        default=os.path.join(xdg.XDG_STATE_HOME, "osc", "cookiejar"),
        description=textwrap.dedent(
            """
            Path to a cookie jar that stores session cookies.
            """
        ),
    )  # type: ignore[assignment]

    section_scm: str = Field(
        default="SCM options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    realname: Optional[str] = Field(
        default=None,
        description=HostOptions.__fields__["realname"].description,
    )  # type: ignore[assignment]

    email: Optional[str] = Field(
        default=None,
        description=HostOptions.__fields__["email"].description,
    )  # type: ignore[assignment]

    local_service_run: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Run local services during commit.
            """
        ),
    )  # type: ignore[assignment]

    getpac_default_project: str = Field(
        default="openSUSE:Factory",
        description=textwrap.dedent(
            """
            The default project for ``osc getpac`` and ``osc bco``.
            The value is a space separated list of strings.
            """
        ),
    )  # type: ignore[assignment]

    exclude_glob: List[str] = Field(
        default=[".osc", "CVS", ".svn", ".*", "_linkerror", "*~", "#*#", "*.orig", "*.bak", "*.changes.vctmp.*"],
        description=textwrap.dedent(
            """
            Space separated list of files ignored by SCM.
            The files can contain globs.
            """
        ),
    )  # type: ignore[assignment]

    exclude_files: List[str] = Field(
        default=[],
        description=textwrap.dedent(
            """
            Files that match the listed glob patterns get skipped during checkout.
            """
        ),
    )  # type: ignore[assignment]

    include_files: List[str] = Field(
        default=[],
        description=textwrap.dedent(
            """
            Files that do not match the listed glob patterns get skipped during checkout.
            The ``exclude_files`` option takes priority over ``include_files``.
            """
        ),
    )  # type: ignore[assignment]

    checkout_no_colon: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Use '/' as project separator instead the default ':' and create corresponding subdirs.
            If enabled, it takes priority over the ``project_separator`` option.
            """
        ),
    )  # type: ignore[assignment]

    project_separator: str = Field(
        default=":",
        description=textwrap.dedent(
            """
            Use the specified string to separate projects.
            """
        ),
    )  # type: ignore[assignment]

    check_filelist: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Check for untracked files and removed files before commit.
            """
        ),
    )  # type: ignore[assignment]

    do_package_tracking: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Track packages in parent project's .osc/_packages.
            """
        ),
    )  # type: ignore[assignment]

    checkout_rooted: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Prevent checking out projects inside other projects or packages.
            """
        ),
    )  # type: ignore[assignment]

    status_mtime_heuristic: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Consider a file with a modified mtime as modified.
            """
        ),
    )  # type: ignore[assignment]

    linkcontrol: bool = Field(
        default=False,
        description=textwrap.dedent(
            # TODO: explain what linkcontrol does
            """
            """
        ),
    )  # type: ignore[assignment]

    section_build: str = Field(
        default="Build options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    build_repository: str = Field(
        default="openSUSE_Factory",
        description=textwrap.dedent(
            """
            The default repository used when the ``repository`` argument is omitted from ``osc build``.
            """
        ),
    )  # type: ignore[assignment]

    buildlog_strip_time: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Strip the build time from the build logs.
            """
        ),
    )  # type: ignore[assignment]

    package_cache_dir: str = Field(
        default="/var/tmp/osbuild-packagecache",
        description=textwrap.dedent(
            """
            The directory where downloaded packages are stored. Must be writable by you.
            """
        ),
        ini_key="packagecachedir",
    )  # type: ignore[assignment]

    no_verify: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Disable signature verification of packages used for build.
            """
        ),
    )  # type: ignore[assignment]

    builtin_signature_check: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Use the RPM's built-in package signature verification.
            """
        ),
    )  # type: ignore[assignment]

    disable_hdrmd5_check: bool = Field(
        default=False,
        description=HostOptions.__fields__["disable_hdrmd5_check"].description,
    )  # type: ignore[assignment]

    section_request: str = Field(
        default="Request options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    include_request_from_project: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            When querying requests, show also those that originate in the specified projects.
            """
        ),
    )  # type: ignore[assignment]

    request_list_days: int = Field(
        default=0,
        description=textwrap.dedent(
            """
            Limit the age of requests shown with ``osc req list`` to the given number of days.

            This is only the default that can be overridden with ``osc request list -D <VALUE>``.
            Use ``0`` for unlimited.
            """
        ),
    )  # type: ignore[assignment]

    check_for_request_on_action: bool = Field(
        default=True,
        description=textwrap.dedent(
            """
            Check for pending requests after executing an action (e.g. checkout, update, commit).
            """
        ),
    )  # type: ignore[assignment]

    request_show_interactive: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Show requests in the interactive mode by default.
            """
        ),
    )  # type: ignore[assignment]

    print_web_links: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Print links to Web UI that can be directly pasted to a web browser where possible.
            """
        ),
    )  # type: ignore[assignment]

    request_show_source_buildstatus: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Print the buildstatus of the source package.
            Works only with ``osc request show`` and the interactive review.
            """
        ),
    )  # type: ignore[assignment]

    submitrequest_accepted_template: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Template message for accepting a request.

            Supported substitutions: ``%(reqid)s``, ``%(type)s``, ``%(who)s``, ``%(src_project)s``, ``%(src_package)s``, ``%(src_rev)s``, ``%(tgt_project)s``, ``%(tgt_package)s``

            Example::

                Hi %(who)s, your request %(reqid)s (type: %(type)s) for %(tgt_project)s/%(tgt_package)s has been accepted. Thank you for your contribution.
            """
        ),
    )  # type: ignore[assignment]

    submitrequest_declined_template: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Template message for declining a request.

            Supported substitutions: ``%(reqid)s``, ``%(type)s``, ``%(who)s``, ``%(src_project)s``, ``%(src_package)s``, ``%(src_rev)s``, ``%(tgt_project)s``, ``%(tgt_package)s``

            Example::

                Hi %(who)s, your request %(reqid)s (type: %(type)s) for %(tgt_project)s/%(tgt_package)s has been declined because ...
            """
        ),
    )  # type: ignore[assignment]

    request_show_review: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Review requests interactively.
            """
        ),
    )  # type: ignore[assignment]

    review_inherit_group: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            If a review was accepted in interactive mode and a group was specified,
            the review will be accepted for this group.
            """
        ),
    )  # type: ignore[assignment]

    submitrequest_on_accept_action: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            What to do with the source package if the request has been accepted.
            If nothing is specified the API default is used.

            Choices: cleanup, update, noupdate
            """
        ),
    )  # type: ignore[assignment]

    # XXX: let's hide attributes from documentation as it is not clear if anyone uses them and should them change from their defaults
    # section_obs_attributes: str = Field(
    #     default="OBS attributes",
    #     exclude=True,
    #     section=True,
    # )  # type: ignore[assignment]

    maintained_attribute: str = Field(
        default="OBS:Maintained",
    )  # type: ignore[assignment]

    maintenance_attribute: str = Field(
        default="OBS:MaintenanceProject",
    )  # type: ignore[assignment]

    maintained_update_project_attribute: str = Field(
        default="OBS:UpdateProject",
    )  # type: ignore[assignment]

    section_build_tool: str = Field(
        default="Build tool options",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    build_jobs: Optional[int] = Field(
        default=os.cpu_count,
        description=textwrap.dedent(
            """
            The number of parallel processes during the build.
            Defaults to the number of available CPU threads.

            If the value is greater than ``0`` then it is passed as ``--jobs`` to the build tool.
            """
        ),
        ini_key="build-jobs",
    )  # type: ignore[assignment]

    vm_type: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Type of the build environment passed the build tool as the ``--vm-type`` option:

             - <empty>   chroot build
             - kvm       KVM VM build (rootless, needs build-device, build-swap, build-memory)
             - xen       XEN VM build (needs build-device, build-swap, build-memory)
             - qemu      [EXPERIMENTAL] QEMU VM build
             - lxc       [EXPERIMENTAL] LXC build
             - uml
             - zvm
             - openstack
             - ec2
             - docker
             - podman    (rootless)
             - pvm
             - nspawn

            See ``build --help`` for more details about supported options.
            """
        ),
        ini_key="build-type",
    )  # type: ignore[assignment]

    build_memory: Optional[int] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The amount of RAM (in MiB) assigned to a build VM.
            """
        ),
        ini_key="build-memory",
    )  # type: ignore[assignment]

    build_root: str = Field(
        default="/var/tmp/build-root%(dash_user)s/%(repo)s-%(arch)s",
        description=textwrap.dedent(
            """
            Path to the build root directory.

            Supported substitutions: ``%(repo)s``, ``%(arch)s``, ``%(project)s``, ``%(package)s``, ``%(apihost)s``, ``%(user)s``, ``%(dash_user)s``
            where::

                - ``apihost`` is the hostname extracted from the currently used ``apiurl``.
                - ``dash_user`` is the username prefixed with a dash. If ``user`` is empty, ``dash_user`` is also empty.

            NOTE: The configuration holds the original unexpanded string. Call ``osc.build.get_build_root()`` with proper arguments to retrieve an actual path.

            Passed as ``--root <VALUE>`` to the build tool.
            """
        ),
        ini_key="build-root",
    )  # type: ignore[assignment]

    build_shell_after_fail: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Start a shell prompt in the build environment if a build fails.

            Passed as ``--shell-after-fail`` to the build tool.
            """
        ),
        ini_key="build-shell-after-fail",
    )  # type: ignore[assignment]

    build_uid: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Numeric uid:gid to use for the abuild user.
            Neither of the values should be 0.
            This is useful if you are hacking in the buildroot.
            This must be set to the same value if the buildroot is re-used.

            Passed as ``--uid <VALUE>`` to the build tool.
            """
        ),
        ini_key="build-uid",
    )  # type: ignore[assignment]

    build_vm_kernel: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The kernel used in a VM build.
            """
        ),
        ini_key="build-kernel",
    )  # type: ignore[assignment]

    build_vm_initrd: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The initrd used in a VM build.
            """
        ),
        ini_key="build-initrd",
    )  # type: ignore[assignment]

    build_vm_disk: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The disk image used as rootfs in a VM build.

            Passed as ``--vm-disk <VALUE>`` to the build tool.
            """
        ),
        ini_key="build-device",
    )  # type: ignore[assignment]

    build_vm_disk_filesystem: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The file system type of the disk image used as rootfs in a VM build.
            Supported values: ext3 (default), ext4, xfs, reiserfs, btrfs.

            Passed as ``--vm-disk-filesystem <VALUE>`` to the build tool.
            """
        ),
        ini_key="build-vmdisk-filesystem",
    )  # type: ignore[assignment]

    build_vm_disk_size: Optional[int] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The size of the disk image (in MiB) used as rootfs in a VM build.

            Passed as ``--vm-disk-size`` to the build tool.
            """
        ),
        ini_key="build-vmdisk-rootsize",
    )  # type: ignore[assignment]

    build_vm_swap: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Path to the disk image used as a swap for VM builds.

            Passed as ``--swap`` to the build tool.
            """
        ),
        ini_key="build-swap",
    )  # type: ignore[assignment]

    build_vm_swap_size: Optional[int] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The size of the disk image (in MiB) used as swap in a VM build.

            Passed as ``--vm-swap-size`` to the build tool.
            """
        ),
        ini_key="build-vmdisk-swapsize",
    )  # type: ignore[assignment]

    build_vm_user: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            The username of a user used to run QEMU/KVM process.
            """
        ),
        ini_key="build-vm-user",
    )  # type: ignore[assignment]

    icecream: int = Field(
        default=0,
        description=textwrap.dedent(
            """
            Use Icecream distributed compiler.
            The value represents the number of parallel build jobs.

            Passed as ``--icecream <VALUE>`` to the build tool.
            """
        ),
    )  # type: ignore[assignment]

    ccache: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Enable compiler cache (ccache) in build roots.

            Passed as ``--ccache`` to the build tool.
            """
        ),
    )  # type: ignore[assignment]

    sccache: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Enable shared compilation cache (sccache) in build roots. Conflicts with ``ccache``.

            Passed as ``--sccache`` to the build tool.
            """
        ),
    )  # type: ignore[assignment]

    sccache_uri: Optional[str] = Field(
        default=None,
        description=textwrap.dedent(
            """
            Optional URI for sccache storage.

            Supported URIs depend on the sccache configuration.
            The URI allows the following substitutions:

             - ``{pkgname}``: name of the package to be build

            Examples:

             - file:///var/tmp/osbuild-sccache-{pkgname}.tar.lzop
             - file:///var/tmp/osbuild-sccache-{pkgname}.tar
             - redis://127.0.0.1:6379

            Passed as ``--sccache-uri <VALUE>`` to the build tool.
            """
        ),
    )  # type: ignore[assignment]

    no_preinstallimage: bool = Field(
        default=False,
        description=textwrap.dedent(
            """
            Do not use preinstall images to initialize build roots.
            """
        ),
    )  # type: ignore[assignment]

    extra_pkgs: List[str] = Field(
        default=[],
        description=textwrap.dedent(
            """
            Extra packages to install into the build root when building packages locally with ``osc build``.

            This corresponds to ``osc build -x pkg1 -x pkg2 ...``.
            The configured values can be overriden from the command-line with ``-x ''``.

            This global setting may leads to dependency problems when the base distro is not providing the package.
            Therefore using server-side ``cli_debug_packages`` option instead is recommended.

            Passed as ``--extra-packs <VALUE>`` to the build tool.
            """
        ),
        ini_key="extra-pkgs",
    )  # type: ignore[assignment]

    section_programs: str = Field(
        default="Paths to programs",
        exclude=True,
        section=True,
    )  # type: ignore[assignment]

    build_cmd: str = Field(
        default=
            shutil.which("build", path="/usr/bin:/usr/lib/build:/usr/lib/obs-build")
            or shutil.which("obs-build", path="/usr/bin:/usr/lib/build:/usr/lib/obs-build")
            or "/usr/bin/build",
        description=textwrap.dedent(
            """
            Path to the 'build' tool.
            """
        ),
        ini_key="build-cmd",
    )  # type: ignore[assignment]

    download_assets_cmd: str = Field(
        default=
            shutil.which("download_assets", path="/usr/lib/build:/usr/lib/obs-build")
            or "/usr/lib/build/download_assets",
        description=textwrap.dedent(
            """
            Path to the 'download_assets' tool used for downloading assets in SCM/Git based builds.
            """
        ),
        ini_key="download-assets-cmd",
    )  # type: ignore[assignment]

    obs_scm_bridge_cmd: str = Field(
        default=
            shutil.which("obs_scm_bridge", path="/usr/lib/obs/service")
            or "/usr/lib/obs/service/obs_scm_bridge",
        description=textwrap.dedent(
            """
            Path to the 'obs_scm_bridge' tool used for cloning scmsync projects and packages.
            """
        ),
        ini_key="obs-scm-bridge-cmd",
    )  # type: ignore[assignment]

    vc_cmd: str = Field(
        default=shutil.which("vc", path="/usr/lib/build:/usr/lib/obs-build") or "/usr/lib/build/vc",
        description=textwrap.dedent(
            """
            Path to the 'vc' tool.
            """
        ),
        ini_key="vc-cmd",
    )  # type: ignore[assignment]

    su_wrapper: str = Field(
        default="sudo",
        description=textwrap.dedent(
            """
            The wrapper to call build tool as root (sudo, su -, ...).
            If empty, the build tool runs under the current user wich works only with KVM at this moment.
            """
        ),
        ini_key="su-wrapper",
    )  # type: ignore[assignment]


# Generate rst from a model. Use it to generate man page in sphinx.
# This IS NOT a public API.
def _model_to_rst(cls, title=None, description=None, sections=None, output_file=None):
    def header(text, char="-"):
        result = f"{text}\n"
        result += f"{'':{char}^{len(text)}}"
        return result

    def bold(text):
        text = text.replace(r"*", r"\*")
        return f"**{text}**"

    def italic(text):
        text = text.replace(r"*", r"\*")
        return f"*{text}*"

    def get_type(name, field):
        ini_type = field.extra.get("ini_type", None)
        if ini_type:
            return ini_type
        if field.origin_type.__name__ == "list":
            return "space-separated-list"
        return field.origin_type.__name__

    def get_default(name, field):
        if field.default is None:
            return None

        if field.default_is_lazy:
            # lazy default may return different results under different circumstances -> return nothing
            return None

        ini_type = field.extra.get("ini_type", None)
        if ini_type:
            return None

        if isinstance(field.default, FromParent):
            return None

        origin_type = field.origin_type

        if origin_type == bool:
            return str(int(field.default))

        if origin_type == int:
            return str(field.default)

        if origin_type == list:
            if not field.default:
                return None
            default_str = " ".join(field.default)
            return f'"{default_str}"'

        if origin_type == str:
            return f'"{field.default}"'

        # TODO:
        raise Exception(f"{name} {field}, {origin_type}")

    result = []

    if title:
        result.append(header(title, char="="))
        result.append("")

    if description:
        result.append(description)
        result.append("")

    for name, field in cls.__fields__.items():
        extra = field.extra

        is_section_header = extra.get("section", False)
        if is_section_header:
            result.append(header(field.default))
            result.append("")
            continue

        exclude = extra.get("ini_exclude", False) or field.exclude
        exclude |= field.description is None
        if exclude:
            continue

        ini_key = extra.get("ini_key", name)

        x = f"{bold(ini_key)} : {get_type(name, field)}"
        default = get_default(name, field)
        if default:
            x += f" = {italic(default)}"
        result.append(x)
        result.append("")
        desc = extra.get("ini_description", None) or field.description or ""
        for line in desc.splitlines():
            result.append(f"    {line}")
        result.append("")

    sections = sections or {}
    for section_name, section_class in sections.items():
        result.append(header(section_name))
        result.append(_model_to_rst(section_class))

    if output_file:
        with open(output_file, "w", encoding="utf-8") as f:
            f.write("\n".join(result))

    return "\n".join(result)


# being global to this module, this object can be accessed from outside
# it will hold the parsed configuration
config = Options()


general_opts = [field.extra.get("ini_key", field.name) for field in Options.__fields__.values() if not field.exclude]
api_host_options = [field.extra.get("ini_key", field.name) for field in HostOptions.__fields__.values() if not field.exclude]


# HACK: Proxy object that modifies field defaults in the Options class; needed for compatibility with the old DEFAULTS dict; prevents breaking osc-plugin-collab
# This IS NOT a public API.
class Defaults:
    def _get_field(self, name):
        if hasattr(Options, name):
            return getattr(Options, name)

        for i in dir(Options):
            field = getattr(Options, i)
            if field.extra.get("ini_key", None) == name:
                return field

        return None

    def __getitem__(self, name):
        field = self._get_field(name)
        result = field.default
        if field.type is List[str]:
            # return list as a string so we can append another string to it
            return ", ".join(result)
        return result

    def __setitem__(self, name, value):
        obj = Options()
        obj.set_value_from_string(name, value)
        field = self._get_field(name)
        field.default = obj[name]


DEFAULTS = Defaults()


new_conf_template = """
# see oscrc(5) man page for the full list of available options

[general]

# Default URL to the API server.
# Credentials and other `apiurl` specific settings must be configured in a `[$apiurl]` config section.
apiurl=%(apiurl)s

[%(apiurl)s]
# aliases=
# user=
# pass=
# credentials_mgr_class=osc.credentials...
"""


account_not_configured_text = """
Your user account / password are not configured yet.
You will be asked for them below, and they will be stored in
%s for future use.
"""

config_incomplete_text = """

Your configuration file %s is not complete.
Make sure that it has a [general] section.
(You can copy&paste the below. Some commented defaults are shown.)

"""

config_missing_apiurl_text = """
The apiurl \'%s\' does not exist in the config file. Please enter
your credentials for this apiurl.
"""


def sanitize_apiurl(apiurl):
    """
    Sanitize apiurl:
    - add https:// schema if apiurl contains none
    - strip trailing slashes
    """
    return urljoin(*parse_apisrv_url(None, apiurl))


def parse_apisrv_url(scheme, apisrv):
    if apisrv.startswith('http://') or apisrv.startswith('https://'):
        url = apisrv
    elif scheme is not None:
        url = scheme + apisrv
    else:
        url = f"https://{apisrv}"
    scheme, url, path = urlsplit(url)[0:3]
    return scheme, url, path.rstrip('/')


def urljoin(scheme, apisrv, path=''):
    return f"{scheme}://{apisrv}" + path


def is_known_apiurl(url):
    """returns ``True`` if url is a known apiurl"""
    apiurl = sanitize_apiurl(url)
    return apiurl in config['api_host_options']


def extract_known_apiurl(url):
    """
    Return longest prefix of given url that is known apiurl,
    None if there is no known apiurl that is prefix of given url.
    """
    scheme, host, path = parse_apisrv_url(None, url)
    p = path.split('/')
    while p:
        apiurl = urljoin(scheme, host, '/'.join(p))
        if apiurl in config['api_host_options']:
            return apiurl
        p.pop()
    return None


def get_apiurl_api_host_options(apiurl):
    """
    Returns all apihost specific options for the given apiurl, ``None`` if
    no such specific options exist.
    """
    # FIXME: in A Better World (tm) there was a config object which
    # knows this instead of having to extract it from a url where it
    # had been mingled into before.  But this works fine for now.

    apiurl = sanitize_apiurl(apiurl)
    if is_known_apiurl(apiurl):
        return config['api_host_options'][apiurl]
    raise oscerr.ConfigMissingApiurl(f'missing credentials for apiurl: \'{apiurl}\'',
                                     '', apiurl)


def get_apiurl_usr(apiurl):
    """
    returns the user for this host - if this host does not exist in the
    internal api_host_options the default user is returned.
    """
    # FIXME: maybe there should be defaults not just for the user but
    # for all apihost specific options.  The ConfigParser class
    # actually even does this but for some reason we don't use it
    # (yet?).

    try:
        return get_apiurl_api_host_options(apiurl)['user']
    except KeyError:
        print('no specific section found in config file for host of [\'%s\'] - using default user: \'%s\''
              % (apiurl, config['user']), file=sys.stderr)
        return config['user']


def get_configParser(conffile=None, force_read=False):
    """
    Returns an ConfigParser() object. After its first invocation the
    ConfigParser object is stored in a method attribute and this attribute
    is returned unless you pass force_read=True.
    """
    if not conffile:
        conffile = identify_conf()

    conffile = os.path.expanduser(conffile)
    if 'conffile' not in get_configParser.__dict__:
        get_configParser.conffile = conffile
    if force_read or 'cp' not in get_configParser.__dict__ or conffile != get_configParser.conffile:
        get_configParser.cp = OscConfigParser.OscConfigParser()
        get_configParser.cp.read(conffile)
        get_configParser.conffile = conffile
    return get_configParser.cp


def write_config(fname, cp):
    """write new configfile in a safe way"""
    # config file is behind a symlink
    # resolve the symlink and continue writing the config as usual
    if os.path.islink(fname):
        fname = os.path.realpath(fname)

    if os.path.exists(fname) and not os.path.isfile(fname):
        # only write to a regular file
        return

    # create directories to the config file (if they don't exist already)
    fdir = os.path.dirname(fname)
    if fdir:
        try:
            os.makedirs(fdir, mode=0o700)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    with open(f"{fname}.new", 'w') as f:
        cp.write(f, comments=True)
    try:
        os.rename(f"{fname}.new", fname)
        os.chmod(fname, 0o600)
    except:
        if os.path.exists(f"{fname}.new"):
            os.unlink(f"{fname}.new")
        raise


def config_set_option(section, opt, val=None, delete=False, update=True, creds_mgr_descr=None, **kwargs):
    """
    Sets a config option. If val is not specified the current/default value is
    returned. If val is specified, opt is set to val and the new value is returned.
    If an option was modified get_config is called with ``**kwargs`` unless update is set
    to ``False`` (``override_conffile`` defaults to ``config['conffile']``).
    If val is not specified and delete is ``True`` then the option is removed from the
    config/reset to the default value.
    """
    cp = get_configParser(config['conffile'])

    if section != 'general':
        section = config.apiurl_aliases.get(section, section)
        scheme, host, path = \
            parse_apisrv_url(config.get('scheme', 'https'), section)
        section = urljoin(scheme, host, path)

    sections = {}
    for url in cp.sections():
        if url == 'general':
            sections[url] = url
        else:
            scheme, host, path = \
                parse_apisrv_url(config.get('scheme', 'https'), url)
            apiurl = urljoin(scheme, host, path)
            sections[apiurl] = url

    section = sections.get(section.rstrip('/'), section)
    if section not in cp.sections():
        raise oscerr.ConfigError(f'unknown section \'{section}\'', config['conffile'])
    if section == 'general' and opt not in general_opts or \
       section != 'general' and opt not in api_host_options:
        raise oscerr.ConfigError(f'unknown config option \'{opt}\'', config['conffile'])

    if not val and not delete and opt == 'pass' and creds_mgr_descr is not None:
        # change password store
        creds_mgr = _get_credentials_manager(section, cp)
        user = _extract_user_compat(cp, section, creds_mgr)
        val = creds_mgr.get_password(section, user, defer=False)

    run = False
    if val:
        if opt == 'pass':
            creds_mgr = _get_credentials_manager(section, cp)
            user = _extract_user_compat(cp, section, creds_mgr)
            old_pw = creds_mgr.get_password(section, user, defer=False)
            try:
                creds_mgr.delete_password(section, user)
                if creds_mgr_descr:
                    creds_mgr_new = creds_mgr_descr.create(cp)
                else:
                    creds_mgr_new = creds_mgr
                creds_mgr_new.set_password(section, user, val)
                write_config(config['conffile'], cp)
                opt = credentials.AbstractCredentialsManager.config_entry
                old_pw = None
            finally:
                if old_pw is not None:
                    creds_mgr.set_password(section, user, old_pw)
                    # not nice, but needed if the Credentials Manager will change
                    # something in cp
                    write_config(config['conffile'], cp)
        else:
            cp.set(section, opt, val)
            write_config(config['conffile'], cp)
        run = True
    elif delete and (cp.has_option(section, opt) or opt == 'pass'):
        if opt == 'pass':
            creds_mgr = _get_credentials_manager(section, cp)
            user = _extract_user_compat(cp, section, creds_mgr)
            creds_mgr.delete_password(section, user)
        else:
            cp.remove_option(section, opt)
        write_config(config['conffile'], cp)
        run = True
    if run and update:
        kw = {
            'override_conffile': config['conffile'],
            'override_no_keyring': config['use_keyring'],
        }
        kw.update(kwargs)
        get_config(**kw)
    if cp.has_option(section, opt):
        return (opt, cp.get(section, opt, raw=True))
    return (opt, None)


def _extract_user_compat(cp, section, creds_mgr):
    """
    This extracts the user either from the ConfigParser or
    the creds_mgr. Only needed for deprecated Gnome Keyring
    """
    user = cp.get(section, 'user')
    if user is None and hasattr(creds_mgr, 'get_user'):
        user = creds_mgr.get_user(section)
    return user


def write_initial_config(conffile, entries, custom_template='', creds_mgr_descriptor=None):
    """
    write osc's intial configuration file. entries is a dict which contains values
    for the config file (e.g. { 'user' : 'username', 'pass' : 'password' } ).
    custom_template is an optional configuration template.
    """
    conf_template = custom_template or new_conf_template
    config = globals()["config"].dict()
    config.update(entries)
    sio = StringIO(conf_template.strip() % config)
    cp = OscConfigParser.OscConfigParser()
    cp.read_file(sio)
    cp.set(config['apiurl'], 'user', config['user'])
    if creds_mgr_descriptor:
        creds_mgr = creds_mgr_descriptor.create(cp)
    else:
        creds_mgr = _get_credentials_manager(config['apiurl'], cp)
    creds_mgr.set_password(config['apiurl'], config['user'], config['pass'])
    write_config(conffile, cp)


def add_section(filename, url, user, passwd, creds_mgr_descriptor=None, allow_http=None):
    """
    Add a section to config file for new api url.
    """
    global config
    cp = get_configParser(filename)
    try:
        cp.add_section(url)
    except OscConfigParser.configparser.DuplicateSectionError:
        # Section might have existed, but was empty
        pass
    cp.set(url, 'user', user)
    if creds_mgr_descriptor:
        creds_mgr = creds_mgr_descriptor.create(cp)
    else:
        creds_mgr = _get_credentials_manager(url, cp)
    creds_mgr.set_password(url, user, passwd)
    if allow_http:
        cp.set(url, 'allow_http', "1")
    write_config(filename, cp)


def _get_credentials_manager(url, cp):
    if cp.has_option(url, credentials.AbstractCredentialsManager.config_entry):
        creds_mgr = credentials.create_credentials_manager(url, cp)
        if creds_mgr is None:
            msg = f'Unable to instantiate creds mgr (section: {url})'
            conffile = get_configParser.conffile
            raise oscerr.ConfigMissingCredentialsError(msg, conffile, url)
        return creds_mgr
    if config['use_keyring'] and GENERIC_KEYRING:
        return credentials.get_keyring_credentials_manager(cp)
    elif cp.get(url, "passx", fallback=None) is not None:
        return credentials.ObfuscatedConfigFileCredentialsManager(cp, None)
    return credentials.PlaintextConfigFileCredentialsManager(cp, None)


def get_config(override_conffile=None,
               override_apiurl=None,
               override_debug=None,
               override_http_debug=None,
               override_http_full_debug=None,
               override_traceback=None,
               override_post_mortem=None,
               override_quiet=None,
               override_no_keyring=None,
               override_verbose=None,
               overrides=None
               ):
    """
    Configure osc.

    The configuration options are loaded with the following priority:
        1. environment variables: ``OSC_<uppercase_option>`` or ``OSC_<uppercase_host_alias>_<uppercase_host_option>``
        2. override arguments provided to ``get_config()``
        3. oscrc config file
    """

    if overrides:
        overrides = overrides.copy()
    else:
        overrides = {}

    if override_apiurl is not None:
        overrides["apiurl"] = override_apiurl

    if override_debug is not None:
        overrides["debug"] = override_debug

    if override_http_debug is not None:
        overrides["http_debug"] = override_http_debug

    if override_http_full_debug is not None:
        overrides["http_full_debug"] = override_http_full_debug

    if override_traceback is not None:
        overrides["traceback"] = override_traceback

    if override_post_mortem is not None:
        overrides["post_mortem"] = override_post_mortem

    if override_no_keyring is not None:
        overrides["use_keyring"] = not override_no_keyring

    if override_quiet is not None:
        overrides["quiet"] = override_quiet

    if override_verbose is not None:
        overrides["verbose"] = override_verbose

    if override_conffile is not None:
        conffile = override_conffile
    else:
        conffile = identify_conf()

    if conffile in ["", "/dev/null"]:
        cp = OscConfigParser.OscConfigParser()
        cp.add_section("general")
    else:
        conffile = os.path.expanduser(conffile)
        if not os.path.exists(conffile):
            raise oscerr.NoConfigfile(conffile, account_not_configured_text % conffile)

        cp = get_configParser(conffile)

        if not cp.has_section("general"):
            # FIXME: it might be sufficient to just assume defaults?
            msg = config_incomplete_text % conffile
            defaults = Options().dict()
            msg += new_conf_template % defaults
            raise oscerr.ConfigError(msg, conffile)

        has_password = False
        for section in cp.sections():
            keys = ["pass", "passx"]
            for key in keys:
                value = cp.get(section, key, fallback="").strip()
                if value:
                    has_password = True
                    break

        # make sure oscrc is not world readable, it may contain a password
        conffile_stat = os.stat(conffile)
        # applying 0o7777 mask because we want to ignore the file type bits
        if conffile_stat.st_mode & 0o7777 != 0o600:
            try:
                os.chmod(conffile, 0o600)
            except OSError as e:
                if e.errno in (errno.EROFS, errno.EPERM):
                    if has_password:
                        print(f"Warning: Configuration file '{conffile}' may have insecure file permissions.", file=sys.stderr)
                else:
                    raise e


    global config

    config = Options()
    config.conffile = conffile

    # read 'debug' value before it gets properly stored into Options for early debug messages
    if override_debug:
        debug_str = str(override_debug)
    elif "OSC_DEBUG" in os.environ:
        debug_str = os.environ["OSC_DEBUG"]
    elif "debug" in cp["general"]:
        debug_str = cp["general"]["debug"]
    else:
        debug_str = "0"
    debug = True if debug_str.strip().lower() in ("1", "yes", "true", "on") else False

    # read host options first in order to populate apiurl aliases
    urls = [i for i in cp.sections() if i != "general"]
    for url in urls:
        apiurl = sanitize_apiurl(url)
        # the username will be overwritten later while reading actual config values
        username = cp[url].get("user", "")
        host_options = HostOptions(apiurl=apiurl, username=username, _parent=config)

        known_ini_keys = set()
        for name, field in host_options.__fields__.items():
            # the following code relies on interating through fields in a given order: aliases, username, credentials_mgr_class, password

            ini_key = field.extra.get("ini_key", name)
            known_ini_keys.add(ini_key)
            known_ini_keys.add(name)

            # iterate through aliases and store the value of the the first env that matches OSC_HOST_{ALIAS}_{NAME}
            env_value = None
            for alias in host_options.aliases:
                alias = alias.replace("-", "_")
                env_key = f"OSC_HOST_{alias.upper()}_{name.upper()}"
                env_value = os.environ.get(env_key, None)
                if env_value is not None:
                    break

            if env_value is not None:
                value = env_value
            elif ini_key in cp[url]:
                value = cp[url][ini_key]
            else:
                value = None

            if name == "credentials_mgr_class":
                # HACK: inject credentials_mgr_class back in case we have specified it from env to have it available for reading password
                if value:
                    cp[url][credentials.AbstractCredentialsManager.config_entry] = value
            elif name == "password":
                creds_mgr = _get_credentials_manager(url, cp)
                if env_value is None:
                    value = creds_mgr.get_password(url, host_options.username, defer=True, apiurl=host_options.apiurl)

            if value is not None:
                host_options.set_value_from_string(name, value)

        for key, value in cp[url].items():
            if key.startswith("_"):
                continue
            if key in known_ini_keys:
                continue
            if debug:
                print(f"DEBUG: Config option '[{url}]/{key}' doesn't map to any HostOptions field", file=sys.stderr)
            host_options[key] = value

        scheme = urlsplit(apiurl)[0]
        if scheme == "http" and not host_options.allow_http:
            msg = "The apiurl '{apiurl}' uses HTTP protocol without any encryption.\n"
            msg += "All communication incl. sending your password IS NOT ENCRYPTED!\n"
            msg += "Add 'allow_http=1' to the [{apiurl}] config file section to mute this message.\n"
            print(msg.format(apiurl=apiurl), file=sys.stderr)

        config.api_host_options[apiurl] = host_options

    # read the main options
    known_ini_keys = set()
    for name, field in config.__fields__.items():
        ini_key = field.extra.get("ini_key", name)
        known_ini_keys.add(ini_key)
        known_ini_keys.add(name)
        env_key = f"OSC_{name.upper()}"

        # priority: env, overrides, config
        if env_key in os.environ:
            value = os.environ[env_key]
            # remove any matching records from overrides because they are checked for emptiness later
            overrides.pop(name, None)
            overrides.pop(ini_key, None)
        elif name in overrides:
            value = overrides.pop(name)
        elif ini_key in overrides:
            value = overrides.pop(ini_key)
        elif ini_key in cp["general"]:
            value = cp["general"][ini_key]
        else:
            continue

        if name == "apiurl":
            # resolve an apiurl alias to an actual apiurl
            apiurl = config.apiurl_aliases.get(value, None)
            if not apiurl:
                # no alias matched, try again with a sanitized apiurl (with https:// prefix)
                # and if there's no match again, just use the sanitized apiurl
                apiurl = sanitize_apiurl(value)
                apiurl = config.apiurl_aliases.get(apiurl, apiurl)
            value = apiurl

        config.set_value_from_string(name, value)

    # BEGIN: override credentials for the default apiurl

    # OSC_APIURL is handled already because it's a regular field
    env_username = os.environ.get("OSC_USERNAME", "")
    env_credentials_mgr_class = os.environ.get("OSC_CREDENTIALS_MGR_CLASS", None)
    env_password = os.environ.get("OSC_PASSWORD", None)

    if config.apiurl not in config.api_host_options:
        host_options = HostOptions(apiurl=config.apiurl, username=env_username, _parent=config)
        config.api_host_options[config.apiurl] = host_options
        # HACK: inject section so we can add credentials_mgr_class later
        cp.add_section(config.apiurl)

    host_options = config.api_host_options[config.apiurl]
    if env_username:
        host_options.set_value_from_string("username", env_username)

    if env_credentials_mgr_class:
        host_options.set_value_from_string("credentials_mgr_class", env_credentials_mgr_class)
        # HACK: inject credentials_mgr_class in case we have specified it from env to have it available for reading password
        cp[config.apiurl]["credentials_mgr_class"] = env_credentials_mgr_class

    if env_password:
        password = Password(env_password)
        host_options.password = password
    elif env_credentials_mgr_class:
        creds_mgr = _get_credentials_manager(config.apiurl, cp)
        password = creds_mgr.get_password(config.apiurl, host_options.username, defer=True, apiurl=host_options.apiurl)
        host_options.password = password

    # END: override credentials for the default apiurl

    for apiurl, host_options in config.api_host_options.items():
        if not host_options.username:
            raise oscerr.ConfigMissingCredentialsError(f"No user configured for apiurl {apiurl}", conffile, apiurl)

        if host_options.password is None:
            raise oscerr.ConfigMissingCredentialsError(f"No password configured for apiurl {apiurl}", conffile, apiurl)

    for key, value in cp["general"].items():
        if key.startswith("_"):
            continue
        if key in known_ini_keys:
            continue
        if debug:
            print(f"DEBUG: Config option '[general]/{key}' doesn't map to any Options field", file=sys.stderr)
        config[key] = value

    if overrides:
        unused_overrides_str = ", ".join((f"'{i}'" for i in overrides))
        raise oscerr.ConfigError(f"Unknown config options: {unused_overrides_str}", "<command-line>")

    # XXX unless config['user'] goes away (and is replaced with a handy function, or
    # config becomes an object, even better), set the global 'user' here as well,
    # provided that there _are_ credentials for the chosen apiurl:
    try:
        config['user'] = get_apiurl_usr(config['apiurl'])
    except oscerr.ConfigMissingApiurl as e:
        e.msg = config_missing_apiurl_text % config['apiurl']
        e.file = conffile
        raise e

    # enable connection debugging after all config options are set
    from .connection import enable_http_debug
    enable_http_debug(config)


def identify_conf():
    # needed for compat reasons(users may have their oscrc still in ~
    if 'OSC_CONFIG' in os.environ:
        return os.environ.get('OSC_CONFIG')

    conffile = os.path.join(xdg.XDG_CONFIG_HOME, "osc", "oscrc")

    if os.path.exists(os.path.expanduser("~/.oscrc")) or os.path.islink(os.path.expanduser("~/.oscrc")):
        if "XDG_CONFIG_HOME" in os.environ:
            print(f"{tty.colorize('WARNING', 'yellow,bold')}: Ignoring XDG_CONFIG_HOME env, loading an existing config from '~/.oscrc' instead", file=sys.stderr)
            print("         To fix this, move the existing '~/.oscrc' to XDG location such as '~/.config/osc/oscrc'", file=sys.stderr)
        elif os.path.exists(os.path.expanduser(conffile)):
            print(f"{tty.colorize('WARNING', 'yellow,bold')}: Ignoring config '{conffile}' in XDG location, loading an existing config from ~/.oscrc instead", file=sys.stderr)
            print("         To fix this, remove '~/.oscrc'", file=sys.stderr)
        return '~/.oscrc'

    return conffile


def interactive_config_setup(conffile, apiurl, initial=True):
    if not apiurl:
        apiurl = Options()["apiurl"]

    scheme = urlsplit(apiurl)[0]
    http = scheme == "http"
    if http:
        msg = "The apiurl '{apiurl}' uses HTTP protocol without any encryption.\n"
        msg += "All communication incl. sending your password WILL NOT BE ENCRYPTED!\n"
        msg += "Do you really want to continue with no encryption?\n"
        print(msg.format(apiurl=apiurl), file=sys.stderr)
        yes = raw_input("Type 'YES' to continue: ")
        if yes != "YES":
            raise oscerr.UserAbort()
        print()

    apiurl_no_scheme = urlsplit(apiurl)[1] or apiurl
    user_prompt = f"Username [{apiurl_no_scheme}]: "
    user = raw_input(user_prompt)
    pass_prompt = f"Password [{user}@{apiurl_no_scheme}]: "
    passwd = getpass.getpass(pass_prompt)
    creds_mgr_descr = select_credentials_manager_descr()
    if initial:
        config = {'user': user, 'pass': passwd}
        if apiurl:
            config['apiurl'] = apiurl
        if http:
            config['allow_http'] = 1
        write_initial_config(conffile, config, creds_mgr_descriptor=creds_mgr_descr)
    else:
        add_section(conffile, apiurl, user, passwd, creds_mgr_descriptor=creds_mgr_descr, allow_http=http)


def select_credentials_manager_descr():
    if not credentials.has_keyring_support():
        print('To use keyrings please install python%d-keyring.' % sys.version_info.major)
    creds_mgr_descriptors = credentials.get_credentials_manager_descriptors()

    rows = []
    for i, creds_mgr_descr in enumerate(creds_mgr_descriptors, 1):
        rows += [str(i), creds_mgr_descr.name(), creds_mgr_descr.description()]

    from .core import build_table
    headline = ('NUM', 'NAME', 'DESCRIPTION')
    table = build_table(len(headline), rows, headline)
    print()
    for row in table:
        print(row)

    i = raw_input('Select credentials manager [default=1]: ')
    if not i:
        i = "1"
    if not i.isdigit():
        sys.exit('Invalid selection')
    i = int(i) - 1
    if i < 0 or i >= len(creds_mgr_descriptors):
        sys.exit('Invalid selection')
    return creds_mgr_descriptors[i]

# vim: sw=4 et



import binascii
import os
import socket
import ssl
import subprocess
import sys
import tempfile
import typing

from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from urllib3.util.ssl_ import create_urllib3_context

from . import oscerr
from .util import xdg


# based on openssl's include/openssl/x509_vfy.h.in
X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT = 18
X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN = 19


def create_ssl_context():
    """
    Create a ssl context with disabled weak crypto.

    Relatively safe defaults are set in urllib3 already,
    but we restrict crypto even more.
    """
    ssl_context = create_urllib3_context()
    # we consider anything older than TLSv1_2 insecure
    if sys.version_info[:2] <= (3, 6):
        # deprecated since py3.7
        ssl_context.options |= ssl.OP_NO_TLSv1
        ssl_context.options |= ssl.OP_NO_TLSv1_1
    else:
        # raise minimum version if too low
        if ssl_context.minimum_version < ssl.TLSVersion.TLSv1_2:
            ssl_context.minimum_version = ssl.TLSVersion.TLSv1_2
    return ssl_context


class CertVerificationError(oscerr.OscBaseError):
    def __str__(self):
        args_str = [str(i) for i in self.args]
        return "Certificate Verification Error: " + "\n".join(args_str)


class TrustedCertStore:
    def __init__(self, ssl_context, host, port):
        self.ssl_context = ssl_context
        self.host = host
        self.port = port or 443

        if not self.host:
            raise ValueError("Empty `host`")

        self.dir_path = os.path.expanduser(os.path.join(xdg.XDG_CONFIG_HOME, "osc", "trusted-certs"))
        if not os.path.isdir(self.dir_path):
            try:
                os.makedirs(self.dir_path, mode=0o700)
            except FileExistsError:
                pass

        file_name = f"{self.host}_{self.port}"
        self.pem_path = os.path.join(self.dir_path, f"{file_name}.pem")
        if os.path.isfile(self.pem_path):
            # load permanently trusted certificate that is stored on disk
            with open(self.pem_path, "rb") as f:
                self.cert = x509.load_pem_x509_certificate(f.read())
            self.ssl_context.load_verify_locations(cafile=self.pem_path)
        else:
            self.cert = None

    def get_server_certificate(self):
        # The following code throws an exception on self-signed certs,
        # therefore we need to retrieve the cert differently.
        # pem = ssl.get_server_certificate((self.host, self.port))

        ssl_context = create_ssl_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
        sock = ssl_context.wrap_socket(socket.socket(), server_hostname=self.host)
        sock.connect((self.host, self.port))
        der = sock.getpeercert(binary_form=True)
        pem = ssl.DER_cert_to_PEM_cert(der)
        cert = x509.load_pem_x509_certificate(pem.encode("utf-8"))
        return cert

    def trust_permanently(self, cert):
        """
        Permanently trust the certificate.
        Store it as a pem file in ~/.config/osc/trusted-certs.
        """
        self.cert = cert
        data = self.cert.public_bytes(serialization.Encoding.PEM)
        with open(self.pem_path, "wb") as f:
            f.write(data)
        self.ssl_context.load_verify_locations(cafile=self.pem_path)

    def trust_temporarily(self, cert):
        """
        Temporarily trust the certificate.
        """
        self.cert = cert
        tmp_dir = os.path.expanduser(os.path.join(xdg.XDG_CONFIG_HOME, "osc"))
        data = self.cert.public_bytes(serialization.Encoding.PEM)
        with tempfile.NamedTemporaryFile(mode="wb+", dir=tmp_dir, prefix="temp_trusted_cert_") as f:
            f.write(data)
            f.flush()
            self.ssl_context.load_verify_locations(cafile=f.name)

    @staticmethod
    def _display_cert(cert):
        print("Subject:", cert.subject.rfc4514_string())
        print("Issuer:", cert.issuer.rfc4514_string())
        try:
            san_ext = cert.extensions.get_extension_for_oid(x509.ExtensionOID.SUBJECT_ALTERNATIVE_NAME)
            san_ext_value = typing.cast(x509.SubjectAlternativeName, san_ext.value)
            san_ext_dnsnames = san_ext_value.get_values_for_type(x509.DNSName)
        except x509.extensions.ExtensionNotFound:
            san_ext_dnsnames = ["(not available)"]
        for san in san_ext_dnsnames:
            print("subjectAltName:", san)
        print("Valid:", cert.not_valid_before, "->", cert.not_valid_after)
        print("Fingerprint(MD5):", binascii.hexlify(cert.fingerprint(hashes.MD5())).decode("utf-8"))
        print("Fingerprint(SHA1):", binascii.hexlify(cert.fingerprint(hashes.SHA1())).decode("utf-8"))

    def prompt_trust(self, cert, reason):
        if self.cert:
            # check if the certificate matches the already trusted certificate for the host and port
            if cert != self.cert:
                raise CertVerificationError([
                    "Remote host identification has changed",
                    "",
                    "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",
                    "IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY!",
                    "",
                    f"Offending certificate is at '{self.pem_path}'"
                ])
        else:
            # since there is no trusted certificate on disk,
            # let's display the server cert and give user options to trust it
            print("The server certificate failed verification")
            print()
            self._display_cert(cert)
            print(f"Reason: {reason}")

            while True:
                print("""
Would you like to
0 - quit (default)
1 - continue anyways
2 - trust the server certificate permanently
9 - review the server certificate
""")

                print("Enter choice [0129]: ", end="")
                r = input()
                if not r or r == "0":
                    raise CertVerificationError(["Untrusted certificate"])
                elif r == "1":
                    self.trust_temporarily(cert)
                    return
                elif r == "2":
                    self.trust_permanently(cert)
                    return
                elif r == "9":
                    # TODO: avoid calling openssl to convert pem to text
                    pem = cert.public_bytes(encoding=serialization.Encoding.PEM).decode("utf-8")
                    cmd = ["openssl", "x509", "-text"]
                    try:
                        cert_text = subprocess.check_output(cmd, input=pem, encoding="utf-8")
                        print(cert_text)
                    except FileNotFoundError:
                        print("ERROR: Unable to display certificate because the 'openssl' executable is not available", file=sys.stderr)



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestSourcediffIssue(XmlModel):
    XML_TAG = "issue"

    state: str = Field(
        xml_attribute=True,
    )

    tracker: str = Field(
        xml_attribute=True,
    )

    name: str = Field(
        xml_attribute=True,
    )

    label: str = Field(
        xml_attribute=True,
    )

    url: str = Field(
        xml_attribute=True,
    )



import os
import unittest

import osc.core
import osc.oscerr

from .common import GET, PUT, OscTestCase


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), 'setlinkrev_fixtures')


def suite():
    return unittest.defaultTestLoader.loadTestsFromTestCase(TestSetLinkRev)


class TestSetLinkRev(OscTestCase):
    def setUp(self):
        super().setUp(copytree=False)

    def _get_fixtures_dir(self):
        return FIXTURES_DIR

    @GET('http://localhost/source/osctest/simple/_link', file='simple_link')
    @GET('http://localhost/source/srcprj/srcpkg?rev=latest', file='simple_filesremote')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Set+link+revision+to+42',
         exp='<link package="srcpkg" project="srcprj" rev="42" />', text='dummytext')
    def test_simple1(self):
        """a simple set_link_rev call without revision"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple')

    @GET('http://localhost/source/osctest/simple/_link', file='simple_link')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Set+link+revision+to+42',
         exp='<link package="srcpkg" project="srcprj" rev="42" />', text='dummytext')
    def test_simple2(self):
        """a simple set_link_rev call with revision"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', '42')

    @GET('http://localhost/source/osctest/simple/_link', file='noproject_link')
    @GET('http://localhost/source/osctest/srcpkg?rev=latest&expand=1', file='expandedsrc_filesremote')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Set+link+revision+to+eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee',
         exp='<link package="srcpkg" rev="eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee" vrev="1" />', text='dummytext')
    def test_expandedsrc(self):
        """expand src package"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', expand=True)

    @GET('http://localhost/source/osctest/simple/_link', file='link_with_rev')
    @GET('http://localhost/source/srcprj/srcpkg?rev=latest', file='simple_filesremote')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Set+link+revision+to+42',
         exp='<link package="srcpkg" project="srcprj" rev="42" />', text='dummytext')
    def test_existingrev(self):
        """link already has a rev attribute, update it to current version"""
        # we could also avoid the superfluous PUT
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple')

    @GET('http://localhost/source/osctest/simple/_link', file='link_with_rev')
    @GET('http://localhost/source/srcprj/srcpkg?rev=latest&expand=1', file='expandedsrc_filesremote')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Set+link+revision+to+eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee',
         exp='<link package="srcpkg" project="srcprj" rev="eeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee" vrev="1" />',
         text='dummytext')
    def test_expandexistingrev(self):
        """link already has a rev attribute, update it to current version"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', expand=True)

    @GET('http://localhost/source/osctest/simple/_link', file='simple_link')
    @GET('http://localhost/source/srcprj/srcpkg?rev=latest&expand=1', text='conflict in file merge', code=400)
    def test_linkerror(self):
        """link is broken"""
        from urllib.error import HTTPError
        # the backend returns status 400 if we try to expand a broken _link
        self.assertRaises(HTTPError, osc.core.set_link_rev, 'http://localhost', 'osctest', 'simple', expand=True)

    @GET('http://localhost/source/osctest/simple/_link', file='rev_link')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Unset+link+revision',
         exp='<link package="srcpkg" project="srcprj" />', text='dummytext')
    def test_deleterev(self):
        """delete rev attribute from link xml"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', revision=None)

    @GET('http://localhost/source/osctest/simple/_link', file='md5_rev_link')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Unset+link+revision',
         exp='<link package="srcpkg" project="srcprj" />', text='dummytext')
    def test_deleterev_md5(self):
        """delete rev and vrev attribute from link xml"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', revision=None)

    @GET('http://localhost/source/osctest/simple/_link', file='simple_link')
    @PUT('http://localhost/source/osctest/simple/_link?comment=Unset+link+revision',
         exp='<link package="srcpkg" project="srcprj" />', text='dummytext')
    def test_deleterevnonexistent(self):
        """delete non existent rev attribute from link xml"""
        osc.core.set_link_rev('http://localhost', 'osctest', 'simple', revision=None)


if __name__ == '__main__':
    unittest.main()



import io
import os
import shutil
import sys
import tempfile
import unittest
from unittest.mock import patch
from urllib.request import HTTPHandler, addinfourl, build_opener
from urllib.parse import urlparse, parse_qs
from xml.etree import ElementTree as ET

import urllib3.response

import osc.conf
import osc.core
from osc.util.xml import xml_fromstring


def urlcompare(url, *args):
    """compare all components of url except query string - it is converted to
    dict, therefor different ordering does not makes url's different, as well
    as quoting of a query string"""

    components = urlparse(url)
    query_args = parse_qs(components.query)
    components = components._replace(query=None)

    if not args:
        return False

    for url in args:
        components2 = urlparse(url)
        query_args2 = parse_qs(components2.query)
        components2 = components2._replace(query=None)

        if components != components2 or \
                query_args != query_args2:
            return False

    return True


def xml_equal(actual, exp):
    try:
        actual_xml = xml_fromstring(actual)
        exp_xml = xml_fromstring(exp)
    except ET.ParseError:
        return False
    todo = [(actual_xml, exp_xml)]
    while todo:
        actual_xml, exp_xml = todo.pop(0)
        if actual_xml.tag != exp_xml.tag:
            return False
        if actual_xml.attrib != exp_xml.attrib:
            return False
        if actual_xml.text != exp_xml.text:
            return False
        if actual_xml.tail != exp_xml.tail:
            return False
        if len(actual_xml) != len(exp_xml):
            return False
        todo.extend(list(zip(actual_xml, exp_xml)))
    return True


class RequestWrongOrder(Exception):
    """raised if an unexpected request is issued to urllib2"""

    def __init__(self, url, exp_url, method, exp_method):
        super().__init__()
        self.url = url
        self.exp_url = exp_url
        self.method = method
        self.exp_method = exp_method

    def __str__(self):
        return '%s, %s, %s, %s' % (self.url, self.exp_url, self.method, self.exp_method)


class RequestDataMismatch(Exception):
    """raised if POSTed or PUTed data doesn't match with the expected data"""

    def __init__(self, url, got, exp):
        self.url = url
        self.got = got
        self.exp = exp

    def __str__(self):
        return '%s, %s, %s' % (self.url, self.got, self.exp)


EXPECTED_REQUESTS = []


# HACK: Fix "ValueError: I/O operation on closed file." error in tests on openSUSE Leap 15.2.
#       The problem seems to appear only in the tests, possibly some interaction with MockHTTPConnectionPool.
#       Porting 753fbc03 to urllib3 in openSUSE Leap 15.2 would fix the problem.
urllib3.response.HTTPResponse.__iter__ = lambda self: iter(self._fp)


class MockHTTPConnectionPool:
    def __init__(self, host, port=None, **conn_kw):
        pass

    def urlopen(self, method, url, body=None, headers=None, retries=None, **response_kw):
        global EXPECTED_REQUESTS
        request = EXPECTED_REQUESTS.pop(0)

        url = f"http://localhost{url}"

        if not urlcompare(request["url"], url) or request["method"] != method:
            raise RequestWrongOrder(request["url"], url, request["method"], method)

        if method in ("POST", "PUT"):
            if 'exp' not in request and 'expfile' in request:
                with open(request['expfile'], 'rb') as f:
                    exp = f.read()
            elif 'exp' in request and 'expfile' not in request:
                exp = request['exp'].encode('utf-8')
            else:
                raise RuntimeError('Specify either `exp` or `expfile`')

            body = body or b""
            if hasattr(body, "read"):
                # if it is a file-like object, read it
                body = body.read()
            if hasattr(body, "encode"):
                # if it can be encoded to bytes, do it
                body = body.encode("utf-8")

            if body != exp:
                # We do not have a notion to explicitly mark xml content. In case
                # of xml, we do not care about the exact xml representation (for
                # now). Hence, if both, data and exp, are xml and are "equal",
                # everything is fine (for now); otherwise, error out
                # (of course, this is problematic if we want to ensure that XML
                # documents are bit identical...)
                if not xml_equal(body, exp):
                    raise RequestDataMismatch(url, repr(body), repr(exp))

        if 'exception' in request:
            raise request["exception"]

        if 'text' not in request and 'file' in request:
            with open(request['file'], 'rb') as f:
                data = f.read()
        elif 'text' in request and 'file' not in request:
            data = request['text'].encode('utf-8')
        else:
            raise RuntimeError('Specify either `text` or `file`')

        response = urllib3.response.HTTPResponse(body=data, status=request.get("code", 200))
        response._fp = io.BytesIO(data)

        return response


def urldecorator(method, url, **kwargs):
    def decorate(test_method):
        def wrapped_test_method(self):
            # put all args into a single dictionary
            kwargs["method"] = method
            kwargs["url"] = url

            # prepend fixtures dir to `file`
            if "file" in kwargs:
                kwargs["file"] = os.path.join(self._get_fixtures_dir(), kwargs["file"])

            # prepend fixtures dir to `expfile`
            if "expfile" in kwargs:
                kwargs["expfile"] = os.path.join(self._get_fixtures_dir(), kwargs["expfile"])

            EXPECTED_REQUESTS.append(kwargs)

            test_method(self)

        # mock connection pool, but only just once
        if not hasattr(test_method, "_MockHTTPConnectionPool"):
            wrapped_test_method = patch('urllib3.HTTPConnectionPool', MockHTTPConnectionPool)(wrapped_test_method)
            wrapped_test_method._MockHTTPConnectionPool = True

        wrapped_test_method.__name__ = test_method.__name__
        return wrapped_test_method
    return decorate


def GET(path, **kwargs):
    return urldecorator('GET', path, **kwargs)


def PUT(path, **kwargs):
    return urldecorator('PUT', path, **kwargs)


def POST(path, **kwargs):
    return urldecorator('POST', path, **kwargs)


def DELETE(path, **kwargs):
    return urldecorator('DELETE', path, **kwargs)


class OscTestCase(unittest.TestCase):
    def setUp(self, copytree=True):
        global EXPECTED_REQUESTS
        EXPECTED_REQUESTS = []
        os.chdir(os.path.dirname(__file__))
        oscrc = os.path.join(self._get_fixtures_dir(), 'oscrc')
        osc.conf.get_config(override_conffile=oscrc, override_no_keyring=True)
        os.environ['OSC_CONFIG'] = oscrc

        self.tmpdir = tempfile.mkdtemp(prefix='osc_test')
        if copytree:
            shutil.copytree(os.path.join(self._get_fixtures_dir(), 'osctest'), os.path.join(self.tmpdir, 'osctest'))
        self.stdout = sys.stdout
        sys.stdout = io.StringIO()

    def tearDown(self):
        sys.stdout = self.stdout
        try:
            shutil.rmtree(self.tmpdir)
        except:
            pass
        os.environ.pop("OSC_CONFIG", "")
        self.assertTrue(len(EXPECTED_REQUESTS) == 0)

    def _get_fixtures_dir(self):
        raise NotImplementedError('subclasses should implement this method')

    def _get_fixture(self, filename):
        path = os.path.join(self._get_fixtures_dir(), filename)
        with open(path) as f:
            return f.read()

    def _change_to_pkg(self, name):
        os.chdir(os.path.join(self.tmpdir, 'osctest', name))

    def _check_list(self, fname, exp):
        fname = os.path.join('.osc', fname)
        self.assertFileContentEqual(fname, exp)

    def _check_addlist(self, exp):
        self._check_list('_to_be_added', exp)

    def _check_deletelist(self, exp):
        self._check_list('_to_be_deleted', exp)

    def _check_conflictlist(self, exp):
        self._check_list('_in_conflict', exp)

    def _check_status(self, p, fname, exp):
        self.assertEqual(p.status(fname), exp)

    def _check_digests(self, fname, *skipfiles):
        fname = os.path.join(self._get_fixtures_dir(), fname)
        with open(os.path.join('.osc', '_files')) as f:
            files_act = f.read()
        with open(fname) as f:
            files_exp = f.read()
        self.assertXMLEqual(files_act, files_exp)
        root = xml_fromstring(files_act)
        for i in root.findall('entry'):
            if i.get('name') in skipfiles:
                continue
            self.assertTrue(os.path.exists(os.path.join('.osc', 'sources', i.get('name'))))
            self.assertEqual(osc.core.dgst(os.path.join('.osc', 'sources', i.get('name'))), i.get('md5'))

    def assertFilesEqual(self, first, second):
        self.assertTrue(os.path.exists(first))
        self.assertTrue(os.path.exists(second))
        with open(first) as f1, open(second) as f2:
            self.assertEqual(f1.read(), f2.read())

    def assertFileContentEqual(self, file_path, expected_content):
        self.assertTrue(os.path.exists(file_path))
        with open(file_path) as f:
            self.assertEqual(f.read(), expected_content)

    def assertFileContentNotEqual(self, file_path, expected_content):
        self.assertTrue(os.path.exists(file_path))
        with open(file_path) as f:
            self.assertNotEqual(f.read(), expected_content)

    def assertXMLEqual(self, act, exp):
        if xml_equal(act, exp):
            return
        # ok, xmls are different, hence, assertEqual is expected to fail
        # (we just use it in order to get a "nice" error message)
        self.assertEqual(act, exp)
        # not reached (unless assertEqual is overridden in an incompatible way)
        raise RuntimeError('assertEqual assumptions violated')

    def assertEqualMultiline(self, got, exp):
        if (got + exp).find('\n') == -1:
            self.assertEqual(got, exp)
        else:
            start_delim = "\n" + (" 8< ".join(["-----"] * 8)) + "\n"
            end_delim = "\n" + (" >8 ".join(["-----"] * 8)) + "\n\n"
            self.assertEqual(got, exp,
                             "got:" + start_delim + got + end_delim +
                             "expected:" + start_delim + exp + end_delim)



import os
import platform
import re
import shlex
import subprocess
import sys
import tempfile
from typing import Dict
from typing import List
from typing import Optional
from typing import TextIO
from typing import Union

from . import tty


def print_msg(*args, print_to: Optional[str] = "debug"):
    """
    Print ``*args`` to the ``print_to`` target:
      - None: print nothing
      - debug: print() to stderr with "DEBUG:" prefix if config["debug"] is set
      - verbose: print() to stdout if config["verbose"] or config["debug"] is set
      - error: print() to stderr with red "ERROR:" prefix
      - warning: print() to stderr with yellow "WARNING:" prefix
      - stdout: print() to stdout
      - stderr: print() to stderr
    """
    from .. import conf

    if print_to is None:
        return
    elif print_to == "debug":
        # print a debug message to stderr if config["debug"] is set
        if conf.config["debug"]:
            print("DEBUG:", *args, file=sys.stderr)
    elif print_to == "verbose":
        # print a verbose message to stdout if config["verbose"] or config["debug"] is set
        if conf.config["verbose"] or conf.config["debug"]:
            print(*args)
    elif print_to == "error":
        print(tty.colorize("ERROR:", "red,bold"), *args, file=sys.stderr)
    elif print_to == "warning":
        print(tty.colorize("WARNING:", "yellow,bold"), *args, file=sys.stderr)
    elif print_to == "stdout":
        # print the message to stdout
        print(*args)
    elif print_to == "stderr":
        # print the message to stderr
        print(*args, file=sys.stderr)
    else:
        raise ValueError(f"Invalid value of the 'print_to' option: {print_to}")


# cached compiled regular expressions; they are created on the first use
SANITIZE_TEXT_RE: Optional[Dict] = None


def sanitize_text(text: Union[bytes, str]) -> Union[bytes, str]:
    """
    Remove forbidden characters and escape sequences from ``text``.

    This must be run on lines or the whole text to work correctly.
    Processing blocks of constant size might lead to splitting escape sequences
    and leaving garbage characters after sanitizing.
    """
    global SANITIZE_TEXT_RE

    if not SANITIZE_TEXT_RE:
        SANITIZE_TEXT_RE = {}

        # CONTROL CHARACTERS
        # remove all control characters with the exception of:
        #   0x09 - horizontal tab (\t)
        #   0x0A - line feed (\n)
        #   0x0D - carriage return (\r)
        #   0x1B - escape - is selectively handled later as part of sanitizing escape sequences

        regex = r"[\x00-\x08\x0B\x0C\x0E-\x1A\x1C-\x1F]"
        SANITIZE_TEXT_RE["str_control"] = re.compile(regex)
        SANITIZE_TEXT_RE["bytes_control"] = re.compile(regex.encode("ascii"))

        # CSI ESCAPE SEQUENCES
        # https://en.wikipedia.org/wiki/ANSI_escape_code#CSI_codes
        # remove all but allowed CSI escape sequences

        # negative lookahead assertion that allows safe color escape sequences
        neg_allowed_csi_sequences = r"(?!\[([0-5]|[34][0-7]|;)+m)"

        # range 0x30–0x3F (OCT \040-\077) (ASCII 0–9:;<=>?); zero or more characters
        csi_parameter_bytes = r"[\x30-\x3F]*"

        # range 0x20–0x2F (OCT \040-\057) (ASCII space and !"#$%&'()*+,-./); zero or more characters
        csi_itermediate_bytes = r"[\x20-\x2F]*"

        # range 0x40–0x7E (OCT \100-\176) (ASCII @A–Z[\]^_`a–z{|}~); 1 character
        csi_final_byte = r"[\x40-\x7E]"

        regex = rf"\033{neg_allowed_csi_sequences}\[{csi_parameter_bytes}{csi_itermediate_bytes}{csi_final_byte}"
        SANITIZE_TEXT_RE["str_csi_sequences"] = re.compile(regex)
        SANITIZE_TEXT_RE["bytes_csi_sequences"] = re.compile(regex.encode("ascii"))

        # FE ESCAPE SEQUENCES
        # https://en.wikipedia.org/wiki/ANSI_escape_code#Fe_Escape_sequences
        # remove all Fe escape sequences

        # range 0x40 to 0x5F (ASCII @A–Z[\]^_); 1 character
        fe = r"[\x40-x5F]"
        regex = rf"\033{neg_allowed_csi_sequences}{fe}"
        SANITIZE_TEXT_RE["str_fe_sequences"] = re.compile(regex)
        SANITIZE_TEXT_RE["bytes_fe_sequences"] = re.compile(regex.encode("ascii"))

        # REMAINING ESCAPE CHARACTERS
        # remove all remaining escape characters that are not followed with the allowed CSI escape sequences

        regex = rf"\033{neg_allowed_csi_sequences}"
        SANITIZE_TEXT_RE["str_esc"] = re.compile(regex)
        SANITIZE_TEXT_RE["bytes_esc"] = re.compile(regex.encode("ascii"))

    if isinstance(text, bytes):
        text = SANITIZE_TEXT_RE["bytes_control"].sub(b"", text)
        text = SANITIZE_TEXT_RE["bytes_csi_sequences"].sub(b"", text)
        text = SANITIZE_TEXT_RE["bytes_fe_sequences"].sub(b"", text)
        text = SANITIZE_TEXT_RE["bytes_esc"].sub(b"", text)
    else:
        text = SANITIZE_TEXT_RE["str_control"].sub("", text)
        text = SANITIZE_TEXT_RE["str_csi_sequences"].sub("", text)
        text = SANITIZE_TEXT_RE["str_fe_sequences"].sub("", text)
        text = SANITIZE_TEXT_RE["str_esc"].sub("", text)
    return text


def safe_print(*args, **kwargs):
    """
    A wrapper to print() that runs sanitize_text() on all arguments.
    """
    args = [sanitize_text(i) for i in args]
    print(*args, **kwargs)


def safe_write(file: TextIO, text: Union[str, bytes], *, add_newline: bool = False):
    """
    Run sanitize_text() on ``text`` and write it to ``file``.

    :param add_newline: Write a newline after writing the ``text``.
    """
    text = sanitize_text(text)
    if isinstance(text, bytes):
        if hasattr(file, "buffer"):
            file.buffer.write(text)
            if add_newline:
                file.buffer.write(os.linesep.encode("utf-8"))
        else:
            # file has no "buffer" attribute, let's try to write the bytes directly
            file.write(text)
            if add_newline:
                file.write(os.linesep.encode("utf-8"))
    else:
        file.write(text)
        if add_newline:
            file.write(os.linesep)


def get_default_pager():
    from ..core import _get_linux_distro

    system = platform.system()
    if system == 'Linux':
        dist = _get_linux_distro()
        if dist == 'debian':
            return 'pager'
        return 'less'
    return 'more'


def get_pager():
    """
    Return (pager, env) where
        ``pager`` is a list with parsed pager command
        ``env`` is copy of os.environ() with added variables specific to the pager
    """
    env = os.environ.copy()
    pager = os.getenv("PAGER", default="").strip()
    pager = pager or get_default_pager()

    # LESS env is not always set and we need -R to display escape sequences properly
    less_opts = os.getenv("LESS", default="")
    if "-R" not in less_opts:
        less_opts += " -R"
    env["LESS"] = less_opts

    return shlex.split(pager), env


def run_pager(message: Union[bytes, str], tmp_suffix: str = ""):
    from ..core import run_external

    if not message:
        return

    if not tty.IS_INTERACTIVE:
        safe_write(sys.stdout, message)
        return

    mode = "w+b" if isinstance(message, bytes) else "w+"
    with tempfile.NamedTemporaryFile(mode=mode, suffix=tmp_suffix) as tmpfile:
        safe_write(tmpfile, message)
        tmpfile.flush()

        pager, env = get_pager()
        cmd = pager + [tmpfile.name]
        run_external(*cmd, env=env)


def pipe_to_pager(lines: Union[List[bytes], List[str]], *, add_newlines=False):
    """
    Pipe ``lines`` to the pager.
    If running in a non-interactive terminal, print the data instead.
    Add a newline after each line if ``add_newlines`` is ``True``.
    """
    if not tty.IS_INTERACTIVE:
        for line in lines:
            safe_write(sys.stdout, line, add_newline=add_newlines)
        return

    pager, env = get_pager()
    with subprocess.Popen(pager, stdin=subprocess.PIPE, encoding="utf-8", env=env) as proc:
        try:
            for line in lines:
                safe_write(proc.stdin, line, add_newline=add_newlines)
                proc.stdin.flush()
            proc.stdin.close()
        except BrokenPipeError:
            pass
        proc.wait()



import hashlib
import os
import shutil
import tempfile
import time
from typing import Optional
from urllib.error import HTTPError
from urllib.parse import urlparse

from .. import oscerr
from .. import output
from ..util.xml import ET


class Serviceinfo:
    """Source service content
    """

    def __init__(self):
        """creates an empty serviceinfo instance"""
        self.services = []
        self.apiurl: Optional[str] = None
        self.project: Optional[str] = None
        self.package: Optional[str] = None

    def read(self, serviceinfo_node, append=False):
        """read in the source services ``<services>`` element passed as
        elementtree node.
        """
        def error(msg, xml):
            from ..core import ET_ENCODING
            data = f'invalid service format:\n{ET.tostring(xml, encoding=ET_ENCODING)}'
            raise ValueError(f"{data}\n\n{msg}")

        if serviceinfo_node is None:
            return
        if not append:
            self.services = []
        services = serviceinfo_node.findall('service')

        for service in services:
            name = service.get('name')
            if name is None:
                error("invalid service definition. Attribute name missing.", service)
            if len(name) < 3 or '/' in name:
                error(f"invalid service name: {name}", service)
            mode = service.get('mode', '')
            data = {'name': name, 'mode': mode}
            command = [name]
            for param in service.findall('param'):
                option = param.get('name')
                if option is None:
                    error(f"{name}: a parameter requires a name", service)
                value = ''
                if param.text:
                    value = param.text
                command.append('--' + option)
                # hmm is this reasonable or do we want to allow real
                # options (e.g., "--force" (without an argument)) as well?
                command.append(value)
            data['command'] = command
            self.services.append(data)

    def getProjectGlobalServices(self, apiurl: str, project: str, package: str):
        from ..core import http_POST
        from ..core import makeurl
        from ..util.xml import xml_parse

        self.apiurl = apiurl
        # get all project wide services in one file, we don't store it yet
        u = makeurl(apiurl, ["source", project, package], query={"cmd": "getprojectservices"})
        try:
            f = http_POST(u)
            root = xml_parse(f).getroot()
            self.read(root, True)
            self.project = project
            self.package = package
        except HTTPError as e:
            if e.code == 404 and package != '_project':
                self.getProjectGlobalServices(apiurl, project, '_project')
                self.package = package
            elif e.code != 403 and e.code != 400:
                raise e

    def addVerifyFile(self, serviceinfo_node, filename: str):
        f = open(filename, 'rb')
        digest = hashlib.sha256(f.read()).hexdigest()
        f.close()

        r = serviceinfo_node
        s = ET.Element("service", name="verify_file")
        ET.SubElement(s, "param", name="file").text = filename
        ET.SubElement(s, "param", name="verifier").text = "sha256"
        ET.SubElement(s, "param", name="checksum").text = digest

        r.append(s)
        return r

    def addDownloadUrl(self, serviceinfo_node, url_string: str):
        url = urlparse(url_string)
        protocol = url.scheme
        host = url.netloc
        path = url.path

        r = serviceinfo_node
        s = ET.Element("service", name="download_url")
        ET.SubElement(s, "param", name="protocol").text = protocol
        ET.SubElement(s, "param", name="host").text = host
        ET.SubElement(s, "param", name="path").text = path

        r.append(s)
        return r

    def addSetVersion(self, serviceinfo_node):
        r = serviceinfo_node
        s = ET.Element("service", name="set_version", mode="buildtime")
        r.append(s)
        return r

    def addGitUrl(self, serviceinfo_node, url_string: Optional[str]):
        r = serviceinfo_node
        s = ET.Element("service", name="obs_scm")
        ET.SubElement(s, "param", name="url").text = url_string
        ET.SubElement(s, "param", name="scm").text = "git"
        r.append(s)
        return r

    def addTarUp(self, serviceinfo_node):
        r = serviceinfo_node
        s = ET.Element("service", name="tar", mode="buildtime")
        r.append(s)
        return r

    def addRecompressTar(self, serviceinfo_node):
        r = serviceinfo_node
        s = ET.Element("service", name="recompress", mode="buildtime")
        ET.SubElement(s, "param", name="file").text = "*.tar"
        ET.SubElement(s, "param", name="compression").text = "xz"
        r.append(s)
        return r

    def execute(self, dir, callmode: Optional[str] = None, singleservice=None, verbose: Optional[bool] = None):
        old_dir = os.path.join(dir, '.old')

        # if 2 osc instances are executed at a time one, of them fails on .old file existence
        # sleep up to 10 seconds until we can create the directory
        for i in reversed(range(10)):
            try:
                os.mkdir(old_dir)
                break
            except FileExistsError:
                time.sleep(1)

            if i == 0:
                msg = f'"{old_dir}" exists, please remove it'
                raise oscerr.OscIOError(None, msg)

        try:
            result = self._execute(dir, old_dir, callmode, singleservice, verbose)
        finally:
            shutil.rmtree(old_dir)
        return result

    def _execute(
        self, dir, old_dir, callmode: Optional[str] = None, singleservice=None, verbose: Optional[bool] = None
    ):
        from ..core import get_osc_version
        from ..core import run_external
        from ..core import vc_export_env

        # cleanup existing generated files
        for filename in os.listdir(dir):
            if filename.startswith('_service:') or filename.startswith('_service_'):
                os.rename(os.path.join(dir, filename),
                          os.path.join(old_dir, filename))

        allservices = self.services or []
        service_names = [s['name'] for s in allservices]
        if singleservice and singleservice not in service_names:
            # set array to the manual specified singleservice, if it is not part of _service file
            data = {'name': singleservice, 'command': [singleservice], 'mode': callmode}
            allservices = [data]
        elif singleservice:
            allservices = [s for s in allservices if s['name'] == singleservice]
            # set the right called mode or the service would be skipped below
            for s in allservices:
                s['mode'] = callmode

        if not allservices:
            # short-circuit to avoid a potential http request in vc_export_env
            # (if there are no services to execute this http request is
            # useless)
            return 0

        # services can detect that they run via osc this way
        os.putenv("OSC_VERSION", get_osc_version())

        # set environment when using OBS 2.3 or later
        if self.project is not None:
            # These need to be kept in sync with bs_service
            os.putenv("OBS_SERVICE_APIURL", self.apiurl)
            os.putenv("OBS_SERVICE_PROJECT", self.project)
            os.putenv("OBS_SERVICE_PACKAGE", self.package)
            # also export vc env vars (some services (like obs_scm) use them)
            vc_export_env(self.apiurl)

        # recreate files
        ret = 0
        for service in allservices:
            if callmode != "all":
                if service['mode'] == "buildtime":
                    continue
                if service['mode'] == "serveronly" and callmode != "local":
                    continue
                if service['mode'] == "manual" and callmode != "manual":
                    continue
                if service['mode'] != "manual" and callmode == "manual":
                    continue
                if service['mode'] == "disabled" and callmode != "disabled":
                    continue
                if service['mode'] != "disabled" and callmode == "disabled":
                    continue
                if service['mode'] != "trylocal" and service['mode'] != "localonly" and callmode == "trylocal":
                    continue
            temp_dir = None
            try:
                temp_dir = tempfile.mkdtemp(dir=dir, suffix=f".{service['name']}.service")
                cmd = service['command']
                if not os.path.exists("/usr/lib/obs/service/" + cmd[0]):
                    raise oscerr.PackageNotInstalled(f"obs-service-{cmd[0]}")
                cmd[0] = "/usr/lib/obs/service/" + cmd[0]
                cmd = cmd + ["--outdir", temp_dir]
                output.print_msg(f"Running source_service '{service['name']}' ...", print_to="stdout")
                output.print_msg("Run source service:", " ".join(cmd), print_to="verbose")
                r = run_external(*cmd)

                if r != 0:
                    print("Aborting: service call failed: ", ' '.join(cmd))
                    # FIXME: addDownloadUrlService calls si.execute after
                    #        updating _services.
                    return r

                if service['mode'] == "manual" or service['mode'] == "disabled" or service['mode'] == "trylocal" or service['mode'] == "localonly" or callmode == "local" or callmode == "trylocal" or callmode == "all":
                    for filename in os.listdir(temp_dir):
                        os.rename(os.path.join(temp_dir, filename), os.path.join(dir, filename))
                else:
                    name = service['name']
                    for filename in os.listdir(temp_dir):
                        os.rename(os.path.join(temp_dir, filename), os.path.join(dir, "_service:" + name + ":" + filename))
            finally:
                if temp_dir is not None:
                    shutil.rmtree(temp_dir)

        return 0



import osc.commandline_git


# we decided not to use the command name 'pull' because that could be confused
# with the completely unrelated 'git pull' command


class PullRequestCommand(osc.commandline_git.GitObsCommand):
    """
    Manage pull requests
    """

    name = "pr"

    def init_arguments(self):
        pass

    def run(self, args):
        self.parser.print_help()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .request_sourcediff_files_file import RequestSourcediffFilesFile
from .request_sourcediff_issue import RequestSourcediffIssue
from .request_sourcediff_new import RequestSourcediffNew
from .request_sourcediff_old import RequestSourcediffOld


class RequestSourcediff(XmlModel):
    XML_TAG = "sourcediff"

    key: str = Field(
        xml_attribute=True,
    )

    old: Optional[RequestSourcediffOld] = Field(
    )

    new: Optional[RequestSourcediffNew] = Field(
    )

    files_list: List[RequestSourcediffFilesFile] = Field(
        xml_name="files",
        xml_wrapped=True,
    )

    issue_list: Optional[List[RequestSourcediffIssue]] = Field(
        xml_name="issues",
        xml_wrapped=True,
        xml_item_name="issue",
    )



# Copyright 2009 Marcus Huewe <suse-tux@gmx.de>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation;
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


import os
import stat
import struct
import sys


# format implementation is based on src/copyin.c and src/util.c (see cpio sources)


class CpioError(Exception):
    """base class for all cpio related errors"""

    def __init__(self, fn, msg):
        super().__init__()
        self.file = fn
        self.msg = msg

    def __str__(self):
        return '%s: %s' % (self.file, self.msg)


class CpioHdr:
    """
    Represents a cpio header ("New" portable format and CRC format).
    """

    def __init__(self, mgc, ino, mode, uid, gid, nlink, mtime, filesize,
                 dev_maj, dev_min, rdev_maj, rdev_min, namesize, checksum,
                 off=-1, filename=b''):
        """
        All passed parameters are hexadecimal strings (not NUL terminated) except
        off and filename. They will be converted into normal ints.
        """
        self.ino = ino
        self.mode = mode
        self.uid = uid
        self.gid = gid
        self.nlink = nlink
        self.mtime = mtime
        # 0 indicates FIFO or dir
        self.filesize = filesize
        self.dev_maj = dev_maj
        self.dev_min = dev_min
        # only needed for special block/char files
        self.rdev_maj = rdev_maj
        self.rdev_min = rdev_min
        # length of filename (inluding terminating NUL)
        self.namesize = namesize
        # != 0 indicates CRC format (which we do not support atm)
        self.checksum = checksum
        for k, v in self.__dict__.items():
            self.__dict__[k] = int(v, 16)
        self.filename = filename
        # data starts at dataoff and ends at dataoff+filesize
        self.dataoff = off

    def __str__(self):
        return "%s %s %s %s" % (self.filename, self.filesize, self.namesize, self.dataoff)


class CpioRead:
    """
    Represents a cpio archive.
    Supported formats:
    * ascii SVR4 no CRC also called "new_ascii"
    """

    # supported formats - use name -> mgc mapping to increase readabilty
    sfmt = {
        'newascii': b'070701',
    }

    # header format
    hdr_fmt = '6s8s8s8s8s8s8s8s8s8s8s8s8s8s'
    hdr_len = 110

    def __init__(self, filename):
        self.filename = filename
        self.format = -1
        self.__file = None
        self._init_datastructs()

    def __del__(self):
        if self.__file:
            self.__file.close()

    def __iter__(self):
        yield from self.hdrs

    def _init_datastructs(self):
        self.hdrs = []

    def _calc_padding(self, off):
        """
        skip some bytes after a header or a file.
        based on 'static void tape_skip_padding()' in copyin.c.
        """
        if self._is_format('newascii'):
            return (4 - (off % 4)) % 4

    def _is_format(self, type):
        return self.format == self.sfmt[type]

    def _copyin_file(self, hdr, dest, fn):
        """saves file to disk"""
        # TODO: investigate links (e.g. symbolic links are working)
        # check if we have a regular file
        if not stat.S_ISREG(stat.S_IFMT(hdr.mode)):
            msg = '\'%s\' is no regular file - only regular files are supported atm' % hdr.filename
            raise NotImplementedError(msg)
        self.__file.seek(hdr.dataoff, os.SEEK_SET)

        if fn.startswith(b"/"):
            raise CpioError(fn, "Extracting files with absolute paths is not supported for security reasons")

        fn = os.path.join(dest, fn)

        dir_path, _ = os.path.split(fn)
        if dir_path:
            os.makedirs(dir_path, exist_ok=True)

        with open(fn, 'wb') as f:
            f.write(self.__file.read(hdr.filesize))
        os.chmod(fn, hdr.mode)
        uid = hdr.uid
        if uid != os.geteuid() or os.geteuid() != 1:
            uid = -1
        gid = hdr.gid
        if gid not in os.getgroups() or os.getegid() != -1:
            gid = -1
        os.chown(fn, uid, gid)

    def _get_hdr(self, fn):
        for h in self.hdrs:
            if h.filename == fn:
                return h
        return None

    def read(self):
        if not self.__file:
            self.__file = open(self.filename, 'rb')
        else:
            self.__file.seek(0, os.SEEK_SET)
        self._init_datastructs()
        data = self.__file.read(6)
        self.format = data
        if self.format not in self.sfmt.values():
            raise CpioError(self.filename, '\'%s\' is not a supported cpio format' % self.format)
        pos = 0
        while len(data) != 0:
            self.__file.seek(pos, os.SEEK_SET)
            data = self.__file.read(self.hdr_len)
            if not data:
                break
            pos += self.hdr_len
            data = struct.unpack(self.hdr_fmt, data)
            hdr = CpioHdr(*data)
            hdr.filename = self.__file.read(hdr.namesize - 1)
            if hdr.filename == b'TRAILER!!!':
                break
            pos += hdr.namesize
            if self._is_format('newascii'):
                pos += self._calc_padding(hdr.namesize + 110)
            hdr.dataoff = pos
            self.hdrs.append(hdr)
            pos += hdr.filesize + self._calc_padding(hdr.filesize)

    def copyin_file(self, filename, dest=None, new_fn=None):
        """
        copies filename to dest.
        If dest is None the file will be stored in $PWD/filename. If dest points
        to a dir the file will be stored in dest/filename. In case new_fn is specified
        the file will be stored as new_fn.
        """

        # accept str for better user experience
        if isinstance(filename, str):
            filename = filename.encode("utf-8")
        if isinstance(dest, str):
            dest = dest.encode("utf-8")
        if isinstance(new_fn, str):
            new_fn = new_fn.encode("utf-8")

        hdr = self._get_hdr(filename)
        if not hdr:
            raise CpioError(filename, '\'%s\' does not exist in archive' % filename)
        dest = dest or os.getcwdb()
        fn = new_fn or filename
        self._copyin_file(hdr, dest, fn)
        return os.path.join(dest, fn).decode("utf-8")

    def copyin(self, dest=None):
        """
        extracts the cpio archive to dest.
        If dest is None $PWD will be used.
        """
        dest = dest or os.getcwdb()
        for h in self.hdrs:
            self._copyin_file(h, dest, h.filename)


class CpioWrite:
    """cpio archive small files in memory, using new style portable header format"""

    def __init__(self):
        self.cpio = bytearray()

    def add(self, name=None, content=None, perms=0x1a4, type=0x8000):
        namesize = len(name) + 1
        if namesize % 2:
            name += b'\0'
        filesize = len(content)
        mode = perms | type

        c = bytearray()
        c.extend(b'070701')  # magic
        c.extend(b'%08X' % 0)  # inode
        c.extend(b'%08X' % mode)  # mode
        c.extend(b'%08X' % 0)  # uid
        c.extend(b'%08X' % 0)  # gid
        c.extend(b'%08X' % 0)  # nlink
        c.extend(b'%08X' % 0)  # mtime
        c.extend(b'%08X' % filesize)
        c.extend(b'%08X' % 0)  # major
        c.extend(b'%08X' % 0)  # minor
        c.extend(b'%08X' % 0)  # rmajor
        c.extend(b'%08X' % 0)  # rminor
        c.extend(b'%08X' % namesize)
        c.extend(b'%08X' % 0)  # checksum

        c.extend(name + b'\0')
        c.extend(b'\0' * (len(c) % 4))

        c.extend(content)

        if len(c) % 4:
            c.extend(b'\0' * (4 - len(c) % 4))

        self.cpio.extend(c)

    def add_padding(self):
        if len(self.cpio) % 512:
            self.cpio.extend(b'\0' * (512 - len(self.cpio) % 512))

    def get(self):
        self.add(b'TRAILER!!!', b'')
        self.add_padding()
        return bytes(self.cpio)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print('usage: %s /path/to/file.cpio' % sys.argv[0])
        sys.exit(1)
    # a potential user might want to pass a bytes instead of a str
    # to make sure that the CpioError's file attribute is always a
    # bytes
    cpio = CpioRead(sys.argv[1])
    cpio.read()
    for hdr in cpio:
        print(hdr)



import re

from .. import oscerr
from .connection import GiteaHTTPResponse


class GiteaException(oscerr.OscBaseError):
    def __init__(self, response: GiteaHTTPResponse):
        self.response = response

    @property
    def status(self):
        return self.response.status

    @property
    def reason(self):
        return self.response.reason

    def __str__(self):
        result = f"{self.status} {self.reason}"
        if self.response.data:
            result += f": {self.response.data}"
        return result


class BranchDoesNotExist(GiteaException):
    def __init__(self, response: GiteaHTTPResponse, owner: str, repo: str, branch: str):
        super().__init__(response)
        self.owner = owner
        self.repo = repo
        self.branch = branch

    def __str__(self):
        result = f"Repo '{self.owner}/{self.repo}' does not contain branch '{self.branch}'"
        return result


class BranchExists(GiteaException):
    def __init__(self, response: GiteaHTTPResponse, owner: str, repo: str, branch: str):
        super().__init__(response)
        self.owner = owner
        self.repo = repo
        self.branch = branch

    def __str__(self):
        result = f"Repo '{self.owner}/{self.repo}' already contains branch '{self.branch}'"
        return result


class ForkExists(GiteaException):
    def __init__(self, response: GiteaHTTPResponse, owner: str, repo: str):
        super().__init__(response)
        self.owner = owner
        self.repo = repo

        regex = re.compile(r".*fork path: (?P<owner>[^/]+)/(?P<repo>[^\]]+)\].*")
        match = regex.match(self.response.json()["message"])
        assert match is not None
        self.fork_owner = match.groupdict()["owner"]
        self.fork_repo = match.groupdict()["repo"]

    def __str__(self):
        result = f"Repo '{self.owner}/{self.repo}' is already forked as '{self.fork_owner}/{self.fork_repo}'"
        return result


class InvalidSshPublicKey(oscerr.OscBaseError):
    def __str__(self):
        return "Invalid public ssh key"



import os
import shutil
import tempfile
import unittest

from osc.util.ar import Ar
from osc.util.ar import ArError


FIXTURES_DIR = os.path.join(os.path.dirname(__file__), "fixtures")


class TestAr(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix="osc_test_")
        try:
            self.old_cwd = os.getcwd()
        except FileNotFoundError:
            self.old_cwd = os.path.expanduser("~")
        os.chdir(self.tmpdir)
        self.archive = os.path.join(FIXTURES_DIR, "archive.ar")
        self.ar = Ar(self.archive)
        self.ar.read()

    def tearDown(self):
        os.chdir(self.old_cwd)
        shutil.rmtree(self.tmpdir)

    def test_file_list(self):
        actual = [i.name for i in self.ar]
        expected = [
            # absolute path
            b"/tmp/foo",
            # this is a filename, not a long filename reference
            b"/123",
            b"very-long-long-long-long-name",
            b"very-long-long-long-long-name2",
            # long file name with a newline
            b"very-long-name\n-with-newline",
            # short file name with a newline
            b"a\nb",
            b"dir/file",
        ]
        self.assertEqual(actual, expected)

    def test_get_file(self):
        f = self.ar.get_file(b"/tmp/foo")
        self.assertIsNotNone(f)

        f = self.ar.get_file("/tmp/foo")
        self.assertIsNotNone(f)

        f = self.ar.get_file("does-not-exist")
        self.assertIsNone(f)

    def test_saveTo(self):
        f = self.ar.get_file("a\nb")
        path = f.saveTo(self.tmpdir)

        # check that we've got the expected path
        self.assertEqual(path, os.path.join(self.tmpdir, "a\nb"))

        # ... and that the contents also match
        with open(path, "r", encoding="utf-8") as f:
            self.assertEqual(f.read(), "newline\n")

    def test_saveTo_subdir(self):
        f = self.ar.get_file("dir/file")
        path = f.saveTo(self.tmpdir)

        # check that we've got the expected path
        self.assertEqual(path, os.path.join(self.tmpdir, "dir/file"))

        # ... and that the contents also match
        with open(path, "r", encoding="utf-8") as f:
            self.assertEqual(f.read(), "file-in-a-dir\n")

    def test_saveTo_abspath(self):
        f = self.ar.get_file("/tmp/foo")
        assert f is not None
        # this is supposed to throw an error, extracting files with absolute paths might overwrite system files
        self.assertRaises(ArError, f.saveTo, self.tmpdir)

    def test_no_exthdr(self):
        self.archive = os.path.join(FIXTURES_DIR, "archive-no-ext_fnhdr.ar")
        self.ar = Ar(self.archive)
        self.ar.read()
        self.test_saveTo_subdir()


if __name__ == "__main__":
    unittest.main()



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from .enums import RequestStates


class RequestState(XmlModel):
    XML_TAG = "state"

    name: RequestStates = Field(
        xml_attribute=True,
    )

    who: Optional[str] = Field(
        xml_attribute=True,
    )

    when: Optional[str] = Field(
        xml_attribute=True,
    )

    created: Optional[str] = Field(
        xml_attribute=True,
    )

    superseded_by: Optional[int] = Field(
        xml_attribute=True,
    )

    approver: Optional[str] = Field(
        xml_attribute=True,
    )

    comment: Optional[str] = Field(
    )



import sys

from .store import GitStore


def warn_experimental():
    print("WARNING: Using EXPERIMENTAL support for git scm. The functionality may change or disappear without a prior notice!", file=sys.stderr)



from ..output.output import print_msg


def format_msg_project_package_options(
    msg,
    project=None,
    package=None,
    dest_project=None,
    dest_package=None,
    repository=None,
    architecture=None,
    dest_repository=None,
    **options,
):
    """
    Format msg, project, package, dest_project, dest_package and options into a meaningful message
    that can be printed out directly or as a debug message.
    """
    if project and not package:
        msg += f" project '{project}'"
    else:
        msg += f" package '{project}/{package}'"

    if repository:
        msg += f" repository '{repository}'"

    if any([dest_project, dest_package, dest_repository]):
        msg += " to"

    if dest_project and not dest_package:
        msg += f" project '{dest_project}'"
    elif dest_project and dest_package:
        msg += f" package '{dest_project}/{dest_package}'"

    if dest_repository:
        msg += f" repository '{dest_repository}'"

    if architecture:
        msg += f" architecture '{architecture}'"

    msg_options = [key.replace("_", "-") for key, value in options.items() if value]
    if msg_options:
        msg_options.sort()
        msg_options_str = ", ".join(msg_options)
        msg += f" options: {msg_options_str}"

    return msg



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RequestActionAcceptinfo(XmlModel):
    XML_TAG = "acceptinfo"

    rev: str = Field(
        xml_attribute=True,
    )

    srcmd5: str = Field(
        xml_attribute=True,
    )

    osrcmd5: str = Field(
        xml_attribute=True,
    )

    oproject: Optional[str] = Field(
        xml_attribute=True,
    )

    opackage: Optional[str] = Field(
        xml_attribute=True,
    )

    xsrcmd5: Optional[str] = Field(
        xml_attribute=True,
    )

    oxsrcmd5: Optional[str] = Field(
        xml_attribute=True,
    )



import subprocess

import osc.commandline_git


class PullRequestCheckoutCommand(osc.commandline_git.GitObsCommand):
    """
    Check out a pull request
    """

    name = "checkout"
    parent = "PullRequestCommand"

    def init_arguments(self):
        self.add_argument(
            "pull",
            type=int,
            help="Number of the pull request",
        )
        self.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Reset the existing local branch to the latest state of the pull request",
        )

    def run(self, args):
        from osc import gitea_api

        self.print_gitea_settings()

        git = gitea_api.Git(".")
        owner, repo = git.get_owner_repo()

        pr = gitea_api.PullRequest.get(self.gitea_conn, owner, repo, args.pull).json()

        head_ssh_url = pr["head"]["repo"]["ssh_url"]
        head_owner = pr["head"]["repo"]["owner"]["login"]
        head_branch = pr["head"]["ref"]

        try:
            git.add_remote(head_owner, head_ssh_url)
        except subprocess.CalledProcessError as e:
            # TODO: check if the remote url matches
            if e.returncode != 3:
                # returncode 3 means that the remote exists; see `man git-remote`
                raise
        git.fetch(head_owner)

        local_branch = git.fetch_pull_request(args.pull, force=args.force)

        # LFS data may not be accessible in the "origin" remote, we need to allow searching in all remotes
        git.set_config("lfs.remote.searchall", "1")

        # configure branch for `git push`
        git.set_config(f"branch.{local_branch}.remote", head_owner)
        git.set_config(f"branch.{local_branch}.pushRemote", head_owner)
        git.set_config(f"branch.{local_branch}.merge", f"refs/heads/{head_branch}")

        # allow `git push` with no arguments to push to a remote branch that is named differently than the local branch
        git.set_config("push.default", "upstream")

        git.switch(local_branch)



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import

from .flag import Flag
from .group_role import GroupRole
from .person_role import PersonRole
from .project_devel import ProjectDevel
from .project_link import ProjectLink
from .project_maintenance_maintains import ProjectMaintenanceMaintains
from .repository import Repository
from .simple_flag import SimpleFlag
from .status import Status


class Project(XmlModel):
    XML_TAG = "project"

    name: str = Field(
        xml_attribute=True,
    )

    class KindEnum(str, Enum):
        STANDARD = "standard"
        MAINTENANCE = "maintenance"
        MAINTENANCE_INCIDENT = "maintenance_incident"
        MAINTENANCE_RELEASE = "maintenance_release"

    kind: Optional[KindEnum] = Field(
        xml_attribute=True,
    )

    title: str = Field(
    )

    description: str = Field(
    )

    url: Optional[str] = Field(
    )

    link_list: Optional[List[ProjectLink]] = Field(
        xml_name="link",
    )

    mountproject: Optional[str] = Field(
    )

    remoteurl: Optional[str] = Field(
    )

    scmsync: Optional[str] = Field(
    )

    devel: Optional[ProjectDevel] = Field(
    )

    person_list: Optional[List[PersonRole]] = Field(
        xml_name="person",
    )

    group_list: Optional[List[GroupRole]] = Field(
        xml_name="group",
    )

    lock: Optional[SimpleFlag] = Field(
        xml_wrapped=True,
    )

    build_list: Optional[List[Flag]] = Field(
        xml_name="build",
        xml_wrapped=True,
    )

    publish_list: Optional[List[Flag]] = Field(
        xml_name="publish",
        xml_wrapped=True,
    )

    useforbuild_list: Optional[List[Flag]] = Field(
        xml_name="useforbuild",
        xml_wrapped=True,
    )

    debuginfo_list: Optional[List[Flag]] = Field(
        xml_name="debuginfo",
        xml_wrapped=True,
    )

    binarydownload_list: Optional[List[Flag]] = Field(
        xml_name="binarydownload",
        xml_wrapped=True,
    )

    sourceaccess: Optional[SimpleFlag] = Field(
        xml_wrapped=True,
    )

    access: Optional[SimpleFlag] = Field(
        xml_wrapped=True,
    )

    maintenance_list: Optional[List[ProjectMaintenanceMaintains]] = Field(
        xml_name="maintenance",
        xml_wrapped=True,
    )

    repository_list: Optional[List[Repository]] = Field(
        xml_name="repository",
    )

    @classmethod
    def from_api(cls, apiurl, project):
        url_path = ["source", project, "_meta"]
        url_query = {}
        response = cls.xml_request("GET", apiurl, url_path, url_query)
        return cls.from_file(response, apiurl=apiurl)

    def to_api(self, apiurl, *, project=None):
        project = project or self.name
        url_path = ["source", project, "_meta"]
        url_query = {}
        response = self.xml_request("PUT", apiurl, url_path, url_query, data=self.to_string())
        return Status.from_file(response, apiurl=apiurl)

    def resolve_repository_flags(self, package_obj=None):
        """
        Resolve the `build`, `debuginfo`, `publish` and `useforbuild` flags
        and return their values for each repository and build arch.

        :returns: {(repo_name, repo_buildarch): {flag_name: bool} for all available repos
        """
        result = {}
        flag_names = ("build", "debuginfo", "publish", "useforbuild")

        # populate the result matrix: {(repo, arch): {"build": None, "debuginfo": None, "publish": None, "useforbuild": None}}
        for repo_obj in self.repository_list or []:
            for arch in repo_obj.arch_list or []:
                result[(repo_obj.name, arch)] = dict([(flag_name, None) for flag_name in flag_names])

        for flag_name in flag_names:
            flag_objects = getattr(self, f"{flag_name}_list") or []
            if package_obj is not None:
                flag_objects += getattr(package_obj, f"{flag_name}_list") or []

            for flag_obj in flag_objects:
                # look up entries matching the current flag and change their values according to the flag's tag
                for (entry_repo, entry_arch), entry_data in result.items():
                    match = flag_obj.repository in (entry_repo, None) and flag_obj.arch in (entry_arch, None)
                    if match:
                        entry_data[flag_name] = True if flag_obj.flag == "enable" else False

        return result



# Copyright (C) 2006 Novell Inc.  All rights reserved.
# This program is free software; it may be used, copied, modified
# and distributed under the terms of the GNU General Public Licence,
# either version 2, or (at your option) any later version.


import glob
import os
import re
import shutil
import subprocess
import sys
import tempfile
from urllib.request import HTTPError

from . import checker as osc_checker
from . import conf
from . import oscerr
from .core import makeurl, dgst
from .grabber import OscFileGrabber, OscMirrorGroup
from .meter import create_text_meter
from .util import packagequery, cpio
from .util.helper import decode_it


class Fetcher:
    def __init__(self, cachedir='/tmp', urllist=None,
                 http_debug=False, cookiejar=None, offline=False,
                 enable_cpio=True, modules=None, download_api_only=False):
        # set up progress bar callback
        self.progress_obj = None
        if sys.stdout.isatty():
            self.progress_obj = create_text_meter(use_pb_fallback=False)

        self.cachedir = cachedir
        # generic download URL lists
        self.urllist = urllist or []
        self.modules = modules or []
        self.http_debug = http_debug
        self.offline = offline
        self.cpio = {}
        self.enable_cpio = enable_cpio
        self.download_api_only = download_api_only

        self.gr = OscFileGrabber(progress_obj=self.progress_obj)

    def __add_cpio(self, pac):
        prpap = f'{pac.project}/{pac.repository}/{pac.repoarch}/{pac.repopackage}'
        self.cpio.setdefault(prpap, {})[pac.repofilename] = pac

    def __download_cpio_archive(self, apiurl, project, repo, arch, package, **pkgs):
        if not pkgs:
            return
        query = {}
        query["binary"] = pkgs
        query["view"] = "cpio"
        query["module"] = self.modules
        try:
            url = makeurl(apiurl, ['build', project, repo, arch, package], query=query)
            sys.stdout.write("preparing download ...\r")
            sys.stdout.flush()
            with tempfile.NamedTemporaryFile(prefix='osc_build_cpio') as tmparchive:
                self.gr.urlgrab(url, filename=tmparchive.name,
                                text=f'fetching packages for \'{project}\'')
                archive = cpio.CpioRead(tmparchive.name)
                archive.read()
                for hdr in archive:
                    # XXX: we won't have an .errors file because we're using
                    # getbinarylist instead of the public/... route
                    # (which is routed to getbinaries)
                    # getbinaries does not support kiwi builds
                    if hdr.filename == b'.errors':
                        archive.copyin_file(hdr.filename)
                        raise oscerr.APIError('CPIO archive is incomplete '
                                              '(see .errors file)')
                    if package == '_repository':
                        n = re.sub(br'\.pkg\.tar\.(zst|.z)$', b'.arch', hdr.filename)
                        if n.startswith(b'container:'):
                            n = re.sub(br'\.tar\.(zst|.z)$', b'.tar', hdr.filename)
                            pac = pkgs[decode_it(n.rsplit(b'.', 1)[0])]
                            pac.canonname = hdr.filename
                        else:
                            pac = pkgs[decode_it(n.rsplit(b'.', 1)[0])]
                    else:
                        # this is a kiwi product
                        pac = pkgs[decode_it(hdr.filename)]

                    # Extract a single file from the cpio archive
                    fd = None
                    tmpfile = None
                    try:
                        fd, tmpfile = tempfile.mkstemp(prefix='osc_build_file')
                        archive.copyin_file(hdr.filename,
                                            decode_it(os.path.dirname(tmpfile)),
                                            decode_it(os.path.basename(tmpfile)))
                        self.move_package(tmpfile, pac.localdir, pac)
                    finally:
                        if fd is not None:
                            os.close(fd)
                        if tmpfile is not None and os.path.exists(tmpfile):
                            os.unlink(tmpfile)

                for pac in pkgs.values():
                    if not os.path.isfile(pac.fullfilename):
                        raise oscerr.APIError('failed to fetch file \'%s\': '
                                              'missing in CPIO archive' %
                                              pac.repofilename)
        except HTTPError as e:
            if e.code != 414:
                raise
            # query str was too large
            keys = list(pkgs.keys())
            if len(keys) == 1:
                raise oscerr.APIError('unable to fetch cpio archive: '
                                      'server always returns code 414')
            n = int(len(pkgs) / 2)
            new_pkgs = {k: pkgs[k] for k in keys[:n]}
            self.__download_cpio_archive(apiurl, project, repo, arch,
                                         package, **new_pkgs)
            new_pkgs = {k: pkgs[k] for k in keys[n:]}
            self.__download_cpio_archive(apiurl, project, repo, arch,
                                         package, **new_pkgs)

    def __fetch_cpio(self, apiurl):
        for prpap, pkgs in self.cpio.items():
            project, repo, arch, package = prpap.split('/', 3)
            self.__download_cpio_archive(apiurl, project, repo, arch, package, **pkgs)

    def fetch(self, pac, prefix=''):
        # for use by the failure callback
        self.curpac = pac

        mg = OscMirrorGroup(self.gr, pac.urllist)

        if self.http_debug:
            print(f'\nURLs to try for package \'{pac}\':', file=sys.stderr)
            print('\n'.join(pac.urllist), file=sys.stderr)
            print(file=sys.stderr)

        try:
            with tempfile.NamedTemporaryFile(prefix='osc_build',
                                             delete=False) as tmpfile:
                mg_stat = mg.urlgrab(pac.filename, filename=tmpfile.name,
                                     text=f'{prefix}({pac.project}) {pac.filename}')
                if mg_stat:
                    self.move_package(tmpfile.name, pac.localdir, pac)

            if not mg_stat:
                if self.enable_cpio:
                    print('%s/%s: attempting download from api, since not found'
                          % (pac.project, pac.name))
                    self.__add_cpio(pac)
                    return
                print()
                print('Error: Failed to retrieve %s from the following locations '
                      '(in order):' % pac.filename, file=sys.stderr)
                print('\n'.join(pac.urllist), file=sys.stderr)
                sys.exit(1)
        finally:
            if os.path.exists(tmpfile.name):
                os.unlink(tmpfile.name)

    def move_package(self, tmpfile, destdir, pac_obj=None):
        canonname = None
        if pac_obj and pac_obj.name.startswith('container:'):
            canonname = pac_obj.canonname
        if canonname is None:
            pkgq = packagequery.PackageQuery.query(tmpfile, extra_rpmtags=(1044, 1051, 1052))
            if pkgq:
                canonname = pkgq.canonname()
            else:
                if pac_obj is None:
                    print('Unsupported file type: ', tmpfile, file=sys.stderr)
                    sys.exit(1)
                canonname = pac_obj.binary
        decoded_canonname = decode_it(canonname)
        if b'/' in canonname or '/' in decoded_canonname:
            raise oscerr.OscIOError(None, 'canonname contains a slash')

        fullfilename = os.path.join(destdir, decoded_canonname)
        if pac_obj is not None:
            pac_obj.canonname = canonname
            pac_obj.fullfilename = fullfilename
        shutil.move(tmpfile, fullfilename)
        os.chmod(fullfilename, 0o644)

    def dirSetup(self, pac):
        dir = os.path.join(self.cachedir, pac.localdir)
        if not os.path.exists(dir):
            try:
                os.makedirs(dir, mode=0o755)
            except OSError as e:
                print('packagecachedir is not writable for you?', file=sys.stderr)
                print(e, file=sys.stderr)
                sys.exit(1)

    def _build_urllist(self, buildinfo, pac):
        if self.download_api_only:
            return []
        urllist = self.urllist
        key = f'{pac.project}/{pac.repository}'
        project_repo_url = buildinfo.urls.get(key)
        if project_repo_url is not None:
            urllist = [project_repo_url]
        return urllist

    def run(self, buildinfo):
        apiurl = buildinfo.apiurl
        cached = 0
        all = len(buildinfo.deps)
        for i in buildinfo.deps:
            urllist = self._build_urllist(buildinfo, i)
            i.makeurls(self.cachedir, urllist)
            # find container extension by looking in the cache
            if i.name.startswith('container:') and i.fullfilename.endswith('.tar.xz'):
                for ext in ['.tar.xz', '.tar.gz', '.tar']:
                    if os.path.exists(i.fullfilename[:-7] + ext):
                        i.canonname = i.canonname[:-7] + ext
                        i.makeurls(self.cachedir, urllist)

            if os.path.exists(i.fullfilename):
                cached += 1
                if not i.name.startswith('container:') and i.pacsuffix != 'rpm':
                    continue

                hdrmd5_is_valid = True
                if i.hdrmd5:
                    if i.name.startswith('container:'):
                        hdrmd5 = dgst(i.fullfilename)
                        if hdrmd5 != i.hdrmd5:
                            hdrmd5_is_valid = False
                    else:
                        hdrmd5 = packagequery.PackageQuery.queryhdrmd5(i.fullfilename)
                        if hdrmd5 != i.hdrmd5:
                            if conf.config["api_host_options"][apiurl]["disable_hdrmd5_check"]:
                                print(f"Warning: Ignoring a hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                                hdrmd5_is_valid = True
                            else:
                                print(f"The file will be redownloaded from the API due to a hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                                hdrmd5_is_valid = False

                    if not hdrmd5_is_valid:
                        os.unlink(i.fullfilename)
                        cached -= 1

        miss = 0
        needed = all - cached
        if all:
            miss = 100.0 * needed / all
        print("%.1f%% cache miss. %d/%d dependencies cached.\n" % (miss, cached, all))
        done = 1
        for i in buildinfo.deps:
            if not os.path.exists(i.fullfilename):
                if self.offline:
                    raise oscerr.OscIOError(None,
                                            'Missing \'%s\' in cache: '
                                            '--offline not possible.' %
                                            i.fullfilename)
                self.dirSetup(i)
                try:
                    # if there isn't a progress bar, there is no output at all
                    prefix = ''
                    if not self.progress_obj:
                        print('%d/%d (%s) %s' % (done, needed, i.project, i.filename))
                    else:
                        prefix = '[%d/%d] ' % (done, needed)
                    self.fetch(i, prefix=prefix)

                    if not os.path.isfile(i.fullfilename):
                        # if the file wasn't downloaded and cannot be found on disk,
                        # mark it for downloading from the API
                        self.__add_cpio(i)
                    else:
                        hdrmd5 = packagequery.PackageQuery.queryhdrmd5(i.fullfilename)
                        if hdrmd5 != i.hdrmd5:
                            if conf.config["api_host_options"][apiurl]["disable_hdrmd5_check"]:
                                print(f"Warning: Ignoring a hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                            else:
                                print(f"The file will be redownloaded from the API due to a hdrmd5 mismatch for {i.fullfilename}: {hdrmd5} (actual) != {i.hdrmd5} (expected)")
                                os.unlink(i.fullfilename)
                                self.__add_cpio(i)

                except KeyboardInterrupt:
                    print('Cancelled by user (ctrl-c)')
                    print('Exiting.')
                    sys.exit(0)
                done += 1

        self.__fetch_cpio(buildinfo.apiurl)

        prjs = list(buildinfo.projects.keys())
        for prj in prjs:
            dest = os.path.join(self.cachedir, prj)
            pubkey_path_base = os.path.join(dest, "_pubkey")
            pubkey_paths = glob.glob(f"{pubkey_path_base}*")

            if self.offline:
                # we're offline, only index the keys found on disk
                if pubkey_paths:
                    for pubkey_path in pubkey_paths:
                        buildinfo.keys.append(pubkey_path)
                    buildinfo.prjkeys.append(prj)
                continue

            from . import obs_api

            os.makedirs(dest, mode=0o755, exist_ok=True)
            pubkeys = []

            try:
                keyinfo = obs_api.Keyinfo.from_api(buildinfo.apiurl, prj)
                for pubkey in keyinfo.pubkey_list or []:
                    pubkeys.append(pubkey.value)
            except HTTPError as e:
                result = obs_api.Keyinfo.get_pubkey_deprecated(buildinfo.apiurl, prj, traverse=True)
                if result:
                    # overwrite ``prj`` with the project that contains the key we're using
                    prj, pubkey = result
                    pubkeys.append(pubkey)

            # remove the existing files, we'll create new files with new contents
            for pubkey_path in pubkey_paths:
                os.unlink(pubkey_path)

            if pubkeys:
                for num, pubkey in enumerate(pubkeys):
                    pubkey_path = f"{pubkey_path_base}-{num}"
                    with open(pubkey_path, "w") as f:
                        f.write(pubkey)
                    buildinfo.keys.append(pubkey_path)
                if prj not in buildinfo.prjkeys:
                    buildinfo.prjkeys.append(prj)


def verify_pacs_old(pac_list):
    """Take a list of rpm filenames and run rpm -K on them.

       In case of failure, exit.

       Check all packages in one go, since this takes only 6 seconds on my Athlon 700
       instead of 20 when calling 'rpm -K' for each of them.
       """
    if not pac_list:
        return

    # don't care about the return value because we check the
    # output anyway, and rpm always writes to stdout.

    # save locale first (we rely on English rpm output here)
    saved_LC_ALL = os.environ.get('LC_ALL')
    os.environ['LC_ALL'] = 'en_EN'

    o = subprocess.Popen(['rpm', '-K'] + pac_list, stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, close_fds=True).stdout

    # restore locale
    if saved_LC_ALL:
        os.environ['LC_ALL'] = saved_LC_ALL
    else:
        os.environ.pop('LC_ALL')

    for line in o.readlines():

        if 'OK' not in line:
            print()
            print('The following package could not be verified:', file=sys.stderr)
            print(line, file=sys.stderr)
            sys.exit(1)

        if 'NOT OK' in line:
            print()
            print('The following package could not be verified:', file=sys.stderr)
            print(line, file=sys.stderr)

            if 'MISSING KEYS' in line:
                missing_key = line.split('#')[-1].split(')')[0]

                print("""
- If the key (%(name)s) is missing, install it first.
  For example, do the following:
    osc signkey PROJECT > file
  and, as root:
    rpm --import %(dir)s/keyfile-%(name)s

  Then, just start the build again.

- If you do not trust the packages, you should configure osc build for XEN or KVM

- You may use --no-verify to skip the verification (which is a risk for your system).
""" % {'name': missing_key,
                    'dir': os.path.expanduser('~')}, file=sys.stderr)

            else:
                print("""
- If the signature is wrong, you may try deleting the package manually
  and re-run this program, so it is fetched again.
""", file=sys.stderr)

            sys.exit(1)


def verify_pacs(bi):
    """Take a list of rpm filenames and verify their signatures.

       In case of failure, exit.
       """

    pac_list = [i.fullfilename for i in bi.deps]
    if conf.config['builtin_signature_check'] is not True:
        return verify_pacs_old(pac_list)

    if not pac_list:
        return

    if not bi.keys:
        raise oscerr.APIError("can't verify packages due to lack of GPG keys")

    print("using keys from", ', '.join(bi.prjkeys))

    failed = False
    checker = osc_checker.Checker()
    try:
        checker.readkeys(bi.keys)
        for pkg in pac_list:
            try:
                checker.check(pkg)
            except Exception as e:
                failed = True
                print(pkg, ':', e)
    except:
        checker.cleanup()
        raise

    if failed:
        checker.cleanup()
        sys.exit(1)

    checker.cleanup()

# vim: sw=4 et



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import


class RepositoryDownloadMaster(XmlModel):
    XML_TAG = "master"

    url: str = Field(
        xml_attribute=True,
    )

    sslfingerprint: Optional[str] = Field(
        xml_attribute=True,
    )



from ..util.models import *  # pylint: disable=wildcard-import,unused-wildcard-import
from xml.etree import ElementTree as ET


class SimpleFlag(XmlModel):
    XML_TAG = None

    def __init__(self, flag, **kwargs):
        super().__init__(flag=flag, **kwargs)

    class SimpleFlagChoices(Enum):
        ENABLE = "enable"
        DISABLE = "disable"

    flag: SimpleFlagChoices = Field(
        xml_set_tag=True,
    )

    def __eq__(self, other):
        if hasattr(other, "flag"):
            return self.flag == other.flag
        # allow comparing with a string
        return self.flag == other

    @classmethod
    def from_xml(cls, root: ET.Element, *, apiurl: Optional[str] = None):
        return cls(flag=root[0].tag)
